"""
Classify cell identities using scNym

scnym_api() is the main API endpoint for users.
This function allows for training and prediction using scnym_train() 
and scnym_predict(). Both of these functions will be infrequently 
accessed by users.

get_pretrained_weights() is a wrapper function that downloads pretrained
weights from our cloud storage bucket.
atlas2target() downloads preprocessed reference datasets and concatenates
them onto a user supplied target dataset.
"""
from typing import Optional, Union, List
from anndata import AnnData
import scanpy as sc
import numpy as np
import pandas as pd
import torch
import os
import os.path as osp
import copy
import pickle
import warnings
import itertools
import pprint
# for fetching pretrained weights, all in standard lib
import requests
import json
import urllib
# for data splits
from sklearn.model_selection import StratifiedKFold
# from scnym
from . import utils
from . import model
from . import main
from . import predict
from . import dataprep

# Define constants

TEST_URL = 'https://storage.googleapis.com/calico-website-mca-storage/kang_2017_stim_pbmc.h5ad'
WEIGHTS_JSON = 'https://storage.googleapis.com/calico-website-scnym-storage/link_tables/pretrained_weights.json'
REFERENCE_JSON = 'https://storage.googleapis.com/calico-website-scnym-storage/link_tables/cell_atlas.json'

ATLAS_ANNOT_KEYS = {
    'human': 'celltype',
    'mouse': 'cell_ontology_class',
    'rat': 'cell_ontology_class',
}

TASKS = (
    'train',
    'predict',
)

# Define configurations

CONFIGS = {
    'default' : {
        'n_epochs': 100,
        'patience': 40,
        'lr': 1.0,
        'optimizer_name': 'adadelta',
        'weight_decay': 1e-4,
        'batch_size': 256,
        'mixup_alpha': 0.3,
        'unsup_max_weight': 1.,
        'unsup_mean_teacher': False,
        'ssl_method': 'mixmatch',
        'ssl_kwargs': {
            'augment_pseudolabels': False,
            'augment': 'log1p_drop',
            'unsup_criterion': 'mse',
            'n_augmentations': 1,
            'T': 0.5,
            'ramp_epochs': 100,
            'burn_in_epochs': 0,
            'dan_criterion': True,
            'dan_ramp_epochs': 20,
            'dan_max_weight': 0.1,
            'min_epochs': 20,
        },
        'model_kwargs' : {
            'n_hidden': 256,
            'n_layers': 2,
            'init_dropout': 0.0,
            'residual': False,
        },
        'tensorboard': False,
    },
}

CONFIGS['no_new_identity'] = copy.deepcopy(CONFIGS['default'])
CONFIGS['no_new_identity']['description'] = (
    'Train scNym models with MixMatch and a domain adversary, assuming no new cell types in the target data.'
)

CONFIGS['new_identity_discovery'] = copy.deepcopy(CONFIGS['default'])
CONFIGS['new_identity_discovery']['ssl_kwargs']['pseudolabel_min_confidence'] = 0.9
CONFIGS['new_identity_discovery']['ssl_kwargs']['dan_use_conf_pseudolabels'] = True
CONFIGS['new_identity_discovery']['description'] = (
    'Train scNym models with MixMatch and a domain adversary, using pseudolabel thresholding to allow for new cell type discoveries.'
)

CONFIGS['no_dan'] = copy.deepcopy(CONFIGS['default'])
CONFIGS['no_dan']['ssl_kwargs']['dan_max_weight'] = 0.0
CONFIGS['no_dan']['ssl_kwargs']['dan_ramp_epochs'] = 1
CONFIGS['no_dan']['description'] = (
    'Train scNym models with MixMatch but no domain adversary. May be useful if class imbalance is very large.'
)

CONFIGS['no_ssl'] = copy.deepcopy(CONFIGS['default'])
CONFIGS['no_ssl']['ssl_kwargs']['dan_max_weight'] = 0.0
CONFIGS['no_ssl']['ssl_kwargs']['dan_ramp_epochs'] = 1
CONFIGS['no_ssl']['ssl_kwargs']['unsup_max_weight'] = 0.0
CONFIGS['no_ssl']['description'] = (
    'Train scNym models with MixMatch but no domain adversary. May be useful if class imbalance is very large.'
)


UNLABELED_TOKEN = 'Unlabeled'


def scnym_api(
    adata: AnnData,
    task: str='train',
    groupby: str=None,
    out_path: str='./scnym_outputs',
    trained_model: str=None,
    config: Union[dict, str]='new_identity_discovery',
    key_added: str='scNym',
    copy: bool=False,
) -> Optional[AnnData]:
    """
    scNym: Semi-supervised adversarial neural networks for 
    single cell classification [Kimmel2020]_.
    
    scNym is a cell identity classifier that transfers annotations from one
    single cell experiment to another. The model is implemented as a neural
    network that employs MixMatch semi-supervision and a domain adversary to
    take advantage of unlabeled data during training. scNym offers superior
    performance to many baseline single cell identity classification methods.
    
    Parameters
    ----------
    adata
        Annotated data matrix used for training or prediction.
        If `"scNym_split"` in `.obs_keys()`, uses the cells annotated 
        `"train", "val"` to select data splits.
    task
        Task to perform, either "train" or "predict".
        If "train", uses `adata` as labeled training data.
        If "predict", uses `trained_model` to infer cell identities for
        observations in `adata`.
    groupby
        Column in `adata.obs` that contains cell identity annotations.
        Values of `"Unlabeled"` indicate that a given cell should be used
        only as unlabeled data during training.
    out_path
        Path to a directory for saving scNym model weights and training logs.
    trained_model
        Path to the output directory of an scNym training run
        or a string specifying a pretrained model.
        If provided while `task == "train"`, used as an initialization. 
    config
        Configuration name or dictionary of configuration of parameters.
        Pre-defined configurations:
            "new_identity_discovery" - Default. Employs pseudolabel thresholding to 
            allow for discovery of new cell identities in the target dataset using
            scNym confidence scores.
            "no_new_identity" - Assumes all cells in the target data belong to one
            of the classes in the training data. Recommended to improve performance
            when this assumption is valid.
    key_added
        Key added to `adata.obs` with scNym predictions if `task=="predict"`.            
    copy
        copy the AnnData object before predicting cell types.
            
    Returns
    -------
    Depending on `copy`, returns or updates `adata` with the following fields.

    `X_scnym` : :class:`~numpy.ndarray`, (:attr:`~anndata.AnnData.obsm`, shape=(n_samples, n_hidden), dtype `float`)
        scNym embedding coordinates of data.
    `scNym` : (`adata.obs`, dtype `str`)
        scNym cell identity predictions for each observation.
    `scNym_train_results` : :class:`~dict`, (:attr:`~anndata.AnnData.uns`)
        results of scNym model training.
    
    Examples
    --------
    >>> import scanpy as sc    
    >>> from scnym.api import scnym_api, atlas2target
    
    **Loading Data and preparing labels**
    
    >>> adata = sc.datasets.kang17()
    >>> target_bidx = adata.obs['stim']=='stim'
    >>> adata.obs['cell'] = np.array(adata.obs['cell'])
    >>> adata.obs.loc[target_bidx, 'cell'] = 'Unlabeled'
    
    **Train an scNym model**
    
    >>> scnym_api(
    ...   adata=adata,
    ...   task='train',
    ...   groupby='clusters',
    ...   out_path='./scnym_outputs',
    ...   config='no_new_identity',
    ... )
    
    **Predict cell identities with the trained scNym model**
    
    >>> path_to_model = './scnym_outputs/'
    >>> scnym_api(
    ...   adata=adata,
    ...   task='predict',
    ...   groupby='scNym',
    ...   trained_model=path_to_model,
    ...   config='no_new_identity',
    ... )
    
    **Perform semi-supervised training with an atlas**
    
    >>> joint_adata = atlas2target(
    ...   adata=adata,
    ...   species='mouse',
    ...   key_added='annotations',
    ... )
    >>> scnym_api(
    ...   adata=joint_adata,
    ...   task='train',
    ...   groupby='annotations',
    ...   out_path='./scnym_outputs',
    ...   config='no_new_identity',
    ... )
    """
    if task not in TASKS:
        msg = f'{task} is not a valid scNym task.\n'
        msg += f'must be one of {TASKS}'
        raise ValueError(msg)
            
    # check configuration arguments and choose a config
    if type(config) == str:
        if config not in CONFIGS.keys():
            msg = f'{config} is not a predefined configuration.\n'
            msg += f'must be one of {CONFIGS.keys()}.'
            raise ValueError(msg)
        else:
            config = CONFIGS[config]
    elif type(config) != dict:
        msg = f'`config` was a {type(config)}, must be dict or str.'
        raise TypeError(msg)
    else:
        # config is a dictionary of parameters
        # add or update default parameters based on these
        dconf = CONFIGS['default']
        for k in config.keys():
            dconf[k] = config[k]
        config = dconf

    # check for CUDA
    if torch.cuda.is_available():
        print('CUDA compute device found.')
    else:
        print('No CUDA device found.')
        print('Computations will be performed on the CPU.')
        print('Add a CUDA compute device to improve speed dramatically.\n')
    
    if not osp.exists(out_path):
        os.makedirs(out_path, exist_ok=True)
    
    # add args to `config`
    config['out_path']  = out_path
    config['groupby']   = groupby
    config['key_added'] = key_added
    config['trained_model'] = trained_model
    
    ################################################
    # check that there are no duplicate genes in the input object
    ################################################
    n_genes = adata.shape[1]
    n_unique_genes = len(np.unique(adata.var_names))
    if n_genes != n_unique_genes:
        msg = 'Duplicate Genes Error\n'
        msg += 'Not all genes passed to scNym were unique.\n'
        msg += f'{n_genes} genes are present but only {n_unique_genes} unique genes were detected.\n'
        msg += 'Please use unique gene names in your input object.\n'
        msg += 'This can be achieved by running `adata.var_names_make_unique()`'
        raise ValueError(msg)
        
    ################################################
    # check that `adata.X` are log1p(CPM) counts
    ################################################
    # we can't directly check if cells were normalized to CPM because
    # users may have filtered out genes *a priori*, so the cell sum
    # may no longer be ~= 1e6.
    # however, we can check that our assumptions about log normalization
    # are true.
    
    # check that the min/max are within log1p(CPM) range
    x_max = np.max(adata.X) > np.log1p(1e6)
    x_min = np.min(adata.X) < 0.
    
    # check to see if a user accidently provided raw counts
    if type(adata.X) == np.ndarray:
        int_counts = np.equal(np.mod(adata.X, 1), 0)
    else:
        int_counts = np.all(np.equal(np.mod(adata.X.data, 1), 0))    
    
    if x_max or x_min or int_counts:
        msg = 'Normalization error\n'
        msg += '`adata.X` does not appear to be log(CountsPerMillion+1) normalized data.\n'
        msg += 'Please replace `adata.X` with log1p(CPM) values.\n'
        msg += '>>> # starting from raw counts in `adata.X`\n'
        msg += '>>> sc.pp.normalize_total(adata, target_sum=1e6))\n'
        msg += '>>> sc.pp.log1p(adata)'
        raise ValueError(msg)
        
    ################################################
    # check inputs and launch the appropriate task
    ################################################
    
    if task == 'train':
        # pass parameters to training routine
        if groupby not in adata.obs.columns:
            msg = f'{groupby} is not a variable in `adata.obs`'
            raise ValueError(msg)
            
        scnym_train(
            adata=adata,
            config=config,
        )
    else:
        # check that a pre-trained model was specified or 
        # provided for prediction
        if trained_model is None:
            msg = 'must provide a path to a trained model for prediction.'
            raise ValueError(msg)
        if not os.path.exists(trained_model) and 'pretrained_' not in trained_model:
            msg = 'path to the trained model does not exist.'
            raise FileNotFoundError(msg)
        # predict identities
        config['model_weights'] = trained_model
        scnym_predict(
            adata=adata,
            config=config,
        )
    return


def scnym_train(
    adata: AnnData,
    config: dict,
) -> None:
    '''Train an scNym model. 
    
    Parameters
    ----------
    adata : AnnData
        [Cells, Genes] experiment containing annotated
        cells to train on.
    config : dict
        configuration options.
    
    Returns
    -------
    None.
    Saves model outputs to `config["out_path"]` and adds model results
    to `adata.uns["scnym_train_results"]`.
    
    Notes
    -----
    This method should only be directly called by advanced users.
    Most users should use `scnym_api`.
    
    See Also
    --------
    scnym_api
    '''
    # determine if unlabeled examples are present
    n_unlabeled = np.sum(
        adata.obs[config['groupby']] == UNLABELED_TOKEN
    )
    if n_unlabeled == 0:
        print('No unlabeled data was found.')
        print(f'Did you forget to set some examples as `"{UNLABELED_TOKEN}"`?')
        print('Proceeding with purely supervised training.')
        print()
        
        unlabeled_counts = None
        unlabeled_genes  = None
        
        X = utils.get_adata_asarray(adata)
        y = pd.Categorical(
            np.array(adata.obs[config['groupby']]),
            categories=np.unique(adata.obs[config['groupby']]),
        ).codes
        class_names = np.unique(adata.obs[config['groupby']])
    else:
        print(f'{n_unlabeled} unlabeled observations found.')
        print('Using unlabeled data as a target set for semi-supervised, adversarial training.')
        print()
        
        target_bidx = adata.obs[config['groupby']] == UNLABELED_TOKEN
        
        train_adata = adata[~target_bidx, :]
        target_adata = adata[target_bidx, :]
        
        print('training examples: ', train_adata.shape)
        print('target   examples: ', target_adata.shape)
        
        X = utils.get_adata_asarray(train_adata)
        y = pd.Categorical(
            np.array(train_adata.obs[config['groupby']]),
            categories=np.unique(train_adata.obs[config['groupby']]),
        ).codes
        unlabeled_counts = utils.get_adata_asarray(target_adata)
        class_names = np.unique(train_adata.obs[config['groupby']])
        
    print('X: ', X.shape)
    print('y: ', y.shape)
    
    if 'scNym_split' not in adata.obs_keys():
        # perform a 90/10 train test split
        traintest_idx = np.random.choice(
            X.shape[0],
            size=int(np.floor(0.9*X.shape[0])),
            replace=False
        )
        val_idx = np.setdiff1d(np.arange(X.shape[0]), traintest_idx)
    else:
        train_idx = np.where(
            train_adata.obs['scNym_split'] == 'train'
        )[0]
        test_idx = np.where(
            train_adata.obs['scNym_split'] == 'test',
        )[0]
        val_idx = np.where(
            train_adata.obs['scNym_split'] == 'val'
        )[0]
        
        if len(train_idx) < 100 or len(test_idx) < 10 or len(val_idx) < 10:
            msg = 'Few samples in user provided data split.\n'
            msg += f'{len(train_idx)} training samples.\n'
            msg += f'{len(test_idx)} testing samples.\n'
            msg += f'{len(val_idx)} validation samples.\n'
            msg += 'Halting.'
            raise RuntimeError(msg)
        # `fit_model()` takes a tuple of `traintest_idx`
        # as a training index and testing index pair.
        traintest_idx = (
            train_idx,
            test_idx,
        )
        
    # check if domain labels were manually specified
    if config.get('domain_groupby', None) is not None:
        domain_groupby = config['domain_groupby']
        # check that the column actually exists
        if domain_groupby not in adata.obs.columns:
            msg = f'no column {domain_groupby} exists in `adata.obs`.\n'
            msg += 'if domain labels are specified, a matching column must exist.'
            raise ValueError(msg)
        # get the label indices as unique integers using pd.Categorical
        # to code each unique label with an int
        domains = np.array(
            pd.Categorical(
                adata.obs[domain_groupby],
                categories=np.unique(adata.obs[domain_groupby]),
            ).codes,
            dtype=np.int32,
        )
        # split domain labels into source and target sets for `fit_model`
        input_domain = domains[~target_bidx]
        unlabeled_domain = domains[target_bidx]
        print('Using user provided domain labels.')
        n_source_doms = len(np.unique(input_domain))
        n_target_doms = len(np.unique(unlabeled_domain))
        print(
            f'Found {n_source_doms} source domains and {n_target_doms} target domains.'
        )
    else:
        # no domains manually supplied, providing `None` to `fit_model`
        # will treat source data as one domain and target data as another
        input_domain = None
        unlabeled_domain = None
        
    # check if pre-trained weights should be used to initialize the model
    if config['trained_model'] is None:
        pretrained = None
    elif 'pretrained_' in config['trained_model']:
        msg = 'pretrained model fetching is not supported for training.'
        raise NotImplementedError(msg)
    else:
        # setup a prediction model
        pretrained = osp.join(
            config['trained_model'],
            '00_best_model_weights.pkl',
        )
        if not osp.exists(pretrained):
            msg = f'{pretrained} file not found.'
            raise FileNotFoundError(msg)
        
    acc, loss = main.fit_model(
        X=X,
        y=y,
        traintest_idx=traintest_idx,
        val_idx=val_idx,
        batch_size=config['batch_size'],
        n_epochs=config['n_epochs'],
        lr=config['lr'],
        optimizer_name=config['optimizer_name'],
        weight_decay=config['weight_decay'],
        ModelClass=model.CellTypeCLF,
        out_path=config['out_path'],
        mixup_alpha=config['mixup_alpha'],
        unlabeled_counts=unlabeled_counts,
        input_domain=input_domain,
        unlabeled_domain=unlabeled_domain,
        unsup_max_weight=config['unsup_max_weight'],
        unsup_mean_teacher=config['unsup_mean_teacher'],
        ssl_method=config['ssl_method'],
        ssl_kwargs=config['ssl_kwargs'],
        pretrained=pretrained,
        patience=config.get('patience', None),
        save_freq=config.get('save_freq', None),
        tensorboard=config.get('tensorboard', False),
        **config['model_kwargs'],
    )
    
    # add the final model results to `adata`
    results = {
        'model_path': osp.realpath(osp.join(config['out_path'], '00_best_model_weights.pkl')),
        'final_acc': acc,
        'final_loss': loss,
        'n_genes': adata.shape[1],
        'n_cell_types': len(np.unique(y)),
        'class_names': class_names,
        'gene_names': adata.var_names.tolist(),
        'model_kwargs': config['model_kwargs'],
        'traintest_idx': traintest_idx,
        'val_idx': val_idx,
    }
    assert osp.exists(results['model_path'])
    
    adata.uns['scNym_train_results'] = results
    
    # save the final model results to disk
    train_results_path = osp.join(
        config['out_path'], 
        'scnym_train_results.pkl',
    )

    with open(train_results_path, 'wb') as f:
        pickle.dump(
            results,
            f
        )
    return


@torch.no_grad()
def scnym_predict(
    adata: AnnData,
    config: dict,
) -> None:
    '''Predict cell identities using an scNym model.
    
    Parameters
    ----------
    adata : AnnData
        [Cells, Genes] experiment containing annotated
        cells to train on.
    config : dict
        configuration options.
        
    Returns
    -------
    None. Adds `adata.obs[config["key_added"]]` and `adata.obsm["X_scnym"]`.
    
    Notes
    -----
    This method should only be directly called by advanced users.
    Most users should use `scnym_api`.
    
    See Also
    --------
    scnym_api    
    '''
    # check if a pretrained model was requested
    if 'pretrained_' in config['trained_model']:
        msg = 'Pretrained Request Error\n'
        msg += 'Pretrained weights are no longer supported in scNym.\n'
        raise NotImplementedError(msg)
#         species = _get_pretrained_weights(
#             trained_model=config['trained_model'],
#             out_path=config['out_path'],
#         )
#         print(f'Successfully downloaded pretrained model for {species}.')
#         config['trained_model'] = config['out_path']
    
    # load training parameters
    with open(
        osp.join(config['trained_model'], 'scnym_train_results.pkl'), 
        'rb',
    ) as f:
        results = pickle.load(f)
        
    # setup a prediction model
    model_weights_path = osp.join(
        config['trained_model'],
        '00_best_model_weights.pkl',
    )
    
    P = predict.Predicter(
        model_weights=model_weights_path,
        n_genes=results['n_genes'],
        n_cell_types=results['n_cell_types'],
        labels=results['class_names'],
        **config['model_kwargs'],
    )
    n_cell_types = results["n_cell_types"]
    n_genes      = results["n_genes"]
    print(f'Loaded model predicting {n_cell_types} classes from {n_genes} features')
    print(results['class_names'])
    
    # Generate a classification matrix
    print('Building a classification matrix...')
    X_raw = utils.get_adata_asarray(adata)
    X = utils.build_classification_matrix(
        X=X_raw,
        model_genes=np.array(results['gene_names']),
        sample_genes=np.array(adata.var_names),
    )
    
    # Predict cell identities
    print('Predicting cell types...')
    pred, names, prob = P.predict(
        X,
        output='prob',
    )
    
    prob = pd.DataFrame(
        prob, 
        columns=results['class_names'], 
        index=adata.obs_names,
    )
    
    # Extract model embeddings
    print('Extracting model embeddings...')
    ds = dataprep.SingleCellDS(X=X, y=np.zeros(X.shape[0]))
    dl = torch.utils.data.DataLoader(
        ds, 
        batch_size=config['batch_size'], 
        shuffle=False,
    )

    model = P.models[0]
    lz_02 = torch.nn.Sequential(
        *list(list(model.modules())[0].children())[1][:-1]
    )
    
    embeddings = []
    for data in dl:
        input_ = data['input']
        input_ = input_.to(device=next(model.parameters()).device)
        z = lz_02(input_)
        embeddings.append(z.detach().cpu())
    Z = torch.cat(embeddings, 0)
    
    # Store results in the anndata object
    adata.obs[config['key_added']] = names
    adata.obs[config['key_added'] + '_confidence'] = np.max(prob, axis=1)
    adata.uns['scNym_probabilities'] = prob
    adata.obsm['X_scnym'] = Z.numpy()
    
    return


def _get_pretrained_weights(
    trained_model: str,
    out_path: str,
) -> str:
    '''Given the name of a set of pretrained model weights,
    fetch weights from GCS and return the model state dict.
    
    Parameters
    ----------
    trained_model : str
        the name of a pretrained model to use, formatted as
        "pretrained_{species}". 
        species should be one of {"human", "mouse", "rat"}.
    out_path : str
        path for saving model weights and outputs.
    
    Returns
    -------
    species : str
        species parsed from the trained model name.
    Saves "{out_path}/00_best_model_weights.pkl" and 
    "{out_path}/scnym_train_results.pkl".
    
    Notes
    -----
    Requires an internet connection to download pre-trained weights.
    '''
    # check that the trained_model argument is valid
    if 'pretrained_' not in trained_model:
        msg = 'pretrained model names must contain `"pretrained_"`'
        raise ValueError(msg)
    
    species = trained_model.split('pretrained_')[1]
    
    # download a table of available pretrained models
    try:
        pretrained_weights_dict = json.loads(
            requests.get(WEIGHTS_JSON).text
        )
    except requests.exceptions.ConnectionError:
        print('Could not download pretrained weighs listing from:')
        print(f'\t{WEIGHTS_JSON}')
        print('Loading pretrained model failed.')

    # check that the species specified has pretrained weights
    if species not in pretrained_weights_dict.keys():
        msg = f'pretrained weights not available for {species}.'
        raise ValueError(species)
    
    # get pretrained weights
    path_for_weights =  osp.join(out_path, f'00_best_model_weights.pkl')
    urllib.request.urlretrieve(
        pretrained_weights_dict[species],
        path_for_weights,
    )
    
    # load model parameters
    model_params = {}
    urllib.request.urlretrieve(
        pretrained_weights_dict['model_params'][species]['gene_names'],
        osp.join(out_path, 'pretrained_gene_names.csv'),
    )
    urllib.request.urlretrieve(
        pretrained_weights_dict['model_params'][species]['class_names'],
        osp.join(out_path, 'pretrained_class_names.csv'),
    )
    model_params['gene_names'] = np.loadtxt(
        osp.join(out_path, 'pretrained_gene_names.csv'),
        delimiter=',',
        dtype='str',
    )
    model_params['class_names'] = np.loadtxt(
        osp.join(out_path, 'pretrained_class_names.csv'),
        delimiter=',',
        dtype='str',
    )
    model_params['n_genes'] = len(model_params['gene_names'])
    model_params['n_cell_types'] = len(model_params['class_names'])
    
    # save model parameters to a results file in the output dir
    path_for_results = f'{out_path}/scnym_train_results.pkl'
    with open(path_for_results, 'wb') as f:
        pickle.dump(model_params, f)
        
    # check that files are present
    if not osp.exists(path_for_weights):
        raise FileNotFoundError(path_for_weights)
    if not osp.exists(path_for_results):
        raise FileNotFoundError(path_for_results)

    return species


def atlas2target(
    adata: AnnData,
    species: str,
    key_added: str='annotations',
) -> AnnData:
    '''Download a preprocessed cell atlas dataset and
    append your new dataset as a target to allow for 
    semi-supervised scNym training.
    
    Parameters
    ----------
    adata : anndata.AnnData
        [Cells, Features] experiment to use as a target
        dataset.
        `adata.var_names` must be formatted as Ensembl gene
        names for the relevant species to match the atlas.
        e.g. `"Gapdh`" for mouse or `"GAPDH"` for human, rather
        than Ensembl gene IDs or another gene annotation.
        
    Returns
    -------
    joint_adata : anndata.AnnData
        [Cells, Features] experiment concatenated with a
        preprocessed cell atlas reference dataset.
        Annotations from the atlas are copied to `.obs[key_added]`
        and all cells in the target dataset `adata` are labeled
        with the special "Unlabeled" token.
        
    Examples
    --------
    >>> adata = sc.datasets.pbmc3k()
    >>> joint_adata = scnym.api.atlas2target(
    ...     adata=adata,
    ...     species='human',
    ...     key_added='annotations',
    ... )
    
    Notes
    -----
    Requires an internet connection to download reference datasets.    
    '''
    # download a directory of cell atlases
    try:
        reference_dict = json.loads(
            requests.get(REFERENCE_JSON).text
        )
    except requests.exceptions.ConnectionError:
        print('Could not download pretrained weighs listing from:')
        print(f'\t{REFERENCE_JSON}')
        print('Loading pretrained model failed.')
    
    # check that the species presented is available
    if species not in reference_dict.keys():
        msg = f'pretrained weights not available for {species}.'
        raise ValueError(species)
        
    # check that there are no gene duplications
    n_uniq_genes = len(np.unique(adata.var_names))
    if n_uniq_genes < len(adata.var_names):
        msg = f'{n_uniq_genes} unique features found, but {adata.shape[1]} features are listed.\n'
        msg += 'Please de-duplicate features in `adata` before joining with an atlas dataset.\n'
        msg += 'Consider `adata.var_names_make_unique()` or aggregating values for features with the same identifier.'
        raise ValueError(msg)

    # download the atlas of interest
    atlas = sc.datasets._datasets.read(
        sc.settings.datasetdir / f'atlas_{species}.h5ad',
        backup_url=reference_dict[species],
    )
    del atlas.raw
    
    # get the key used by the cell atlas
    atlas_annot_key = ATLAS_ANNOT_KEYS[species]
    
    # copy atlas annotations to the specified column
    atlas.obs[key_added] = np.array(atlas.obs[atlas_annot_key])
    atlas.obs['scNym_dataset'] = 'atlas_reference'
    
    # label target data with "Unlabeled"
    adata.obs[key_added] = 'Unlabeled'
    adata.obs['scNym_dataset'] = 'target'
    
    # check that at least some genes overlap between the atlas
    # and the target data
    FEW_GENES = 100
    n_overlapping_genes = len(np.intersect1d(adata.var_names, atlas.var_names))
    if n_overlapping_genes == 0:
        msg = 'No genes overlap between the target data `adata` and the atlas.\n'
        msg += 'Genes in the atlas are named using Ensembl gene symbols (e.g. `"Gapdh"`).\n'
        msg += 'Ensure `adata.var_names` also uses gene symbols.'
        raise RuntimeError(msg)
    elif n_overlapping_genes < FEW_GENES:
        msg = f'Only {n_overlapping_genes} overlapping genes were found between the target and atlas.\n'
        msg += 'Ensure your target dataset `adata.var_names` are Ensembl gene names.\n'
        msg += 'Continuing with transer, but performance is likely to be poor.'
        warnings.warn(msg)
    else:
        msg = f'{n_overlapping_genes} overlapping genes found between the target and atlas data.'
        print(msg)
    
    # join the target and atlas data
    joint_adata = atlas.concatenate(
        adata, 
        join='inner',
    )
        
    return joint_adata


def list_configs():
    for k in CONFIGS.keys():
        print(f'name: {k}')
        print('\t'+CONFIGS[k]['description'])
    return


def _get_keys_and_list(d: dict) -> (List[list], List[list]):
    '''Get a set of keys mapping to a list in a
    nested dictionary structure and the list value.
    
    Parameters
    ----------
    d : dict
        a nested dictionary structure where all terminal
        values are lists.
    
    Returns
    -------
    keys : List[list]
        sequential keys required to access a set of 
        associated terminal values.
        mapped by index to `values`.
    values : List[list]
        lists of terminal values, each accessed by the
        set of `keys` with a matching index from `d`.
    '''
    accession_keys = []
    associated_values = []
    for k in d.keys():
        if type(d[k])==dict:
            # the value is nested, recurse
            keys, values = _get_keys_and_list(d[k])
            keys = [[k,]+x for x in keys]
        else:
            keys = [[k],]
            values = [d[k]]
        
        for i in range(len(values)):
            accession_keys.append(keys[i])
            associated_values.append(values[i])

    return accession_keys, associated_values


def _updated_nested(d: dict, keys: list, value: list) -> dict:
    '''Updated the values in a dictionary with multiple nested levels.
    
    Parameters
    ----------
    d : dict
        multilevel dictionary.
    keys : list
        sequential keys specifying a value to update
    value : list
        new value to use in the update.
    
    Returns
    -------
    d : dict
        updated dictionary.
    '''
    if type(d.get(keys[0], None)) == dict:
        # multilevel, recurse
        _updated_nested(d[keys[0]], keys[1:], value)
    else: 
        d[keys[0]] = value
    return


def split_data(
    adata: AnnData,
    groupby: str,
    n_splits: int,
) -> None:
    '''Split data using a stratified k-fold.
    
    Parameters
    ----------
    adata : anndata.AnnData
        [Cells, Genes] experiment.
    groupby : str
        annotation column in `.obs`.
        used for stratification.
    n_splits : int
        number of train/test/val splits to perform for tuning.
        performs at least 5-fold splitting and uses a subset of
        the folds if `n_splits < 5`.
        
    Returns
    -------
    None. Adds `f"scNym_split_{n}"` to `adata.obs` for all `n`
    in `[0, n_splits)`.
    '''
    # generate cross val splits
    cv = StratifiedKFold(
        n_splits=max(5, n_splits), 
        shuffle=True,
    )
    split_indices = list(
        cv.split(adata.X, adata.obs[groupby])
    )

    for split_number, train_test in enumerate(split_indices):

        train_idx = train_test[0]
        testval_idx = train_test[1]

        test_idx = np.random.choice(
            testval_idx,
            size=int(np.ceil(len(testval_idx)/2)),
            replace=False,
        )
        val_idx = np.setdiff1d(
            testval_idx,
            test_idx,
        )

        # these tokens are recognized by `api.scnym_train`
        adata.obs[f'scNym_split_{split_number}'] = 'ERROR'
        adata.obs.loc[
            adata.obs_names[train_idx], f'scNym_split_{split_number}'
        ] = 'train'
        adata.obs.loc[
            adata.obs_names[test_idx], f'scNym_split_{split_number}'
        ] = 'test'    
        adata.obs.loc[
            adata.obs_names[val_idx], f'scNym_split_{split_number}'
        ] = 'val'

    return


def _circular_train(
    search_config: dict,
    params: tuple,
    adata: AnnData,
    groupby: str,
    out_path: str,
    accession_keys: List[list],
    hold_out_only: bool,
    groupby_eval: str,
) -> pd.DataFrame:
    '''
    Perform a circular training loop for a parameter set.
    
    Parameters
    ----------
    search_config : tuple
        configuration for parameter search.
    params : tuple
        search parameter values
    adata : anndata.AnnData
        [Cells, Genes] experiment for optimization.
    groupby : str
        annotation column in `.obs`.
    accession_keys : List[list]
        sequential keys required to access a set of 
        associated terminal values.
        mapped by index to `values`.
    hold_out_only : bool
        evaluate the circular accuracy only on a held-out set of 
        training data, not used in the training of the first
        source -> target model.        

    Returns
    -------
    search_df : pd.DataFrame
        [1, (params,) + (acc,)]
    search_config : dict
        adjusted configuration file for this parameter search.
    '''
    search_number = search_config['search_number']
    split_number  = search_config['split_number']
    # fit the source2target
    s2t_out_path = osp.join(out_path, f'search_{search_number:04}_split_{split_number:04}_source2target')
    adata = adata.copy()
    
    print('\n>>>\nTraining source2target model\n>>>\n')
    scnym_api(
        adata=adata,
        groupby=groupby,
        task='train',
        out_path=s2t_out_path,
        config=search_config,
    )
    
    # load the hold out test acc
    with open(osp.join(s2t_out_path, 'scnym_train_results.pkl'), 'rb') as f:
        s2t_res = pickle.load(f)
        s2t_source_test_acc = s2t_res['final_acc']

    print('\n>>>\nPredicting with source2target model\n>>>\n')
    # predict on the target set
    scnym_api(
        adata=adata,
        task='predict',
        trained_model=s2t_out_path,
        config=search_config,
    )

    # invert the problem -- train on the new labels
    circ_adata = adata.copy()
    circ_adata.obs[groupby] = adata.obs['scNym']
    circ_adata.obs.drop(columns=['scNym'], inplace=True)
    # set the training data as unlabeled, leaving labels only on the target data
    circ_adata.obs.loc[adata.obs[groupby]!=UNLABELED_TOKEN, groupby] = UNLABELED_TOKEN

    # fit a new model
    t2s_out_path = osp.join(out_path, f'search_{search_number:04}_split_{split_number:04}_target2source')
    
    print('\n>>>\nTraining target2source model\n>>>\n')
    
    scnym_api(
        adata=circ_adata,
        groupby=groupby,
        task='train',
        out_path=t2s_out_path,
        config=search_config,
    )

    # predict with new model
    print('\n>>>\nPredicting with target2source model\n>>>\n')    
    scnym_api(
        adata=circ_adata,
        task='predict',
        trained_model=t2s_out_path,
        config=search_config,
    )

    # evaluate the model
    samples_bidx = adata.obs[groupby]!='Unlabeled'
    samples_bidx = (
        samples_bidx & (adata.obs['scNym_split']=='val') if hold_out_only else samples_bidx
    )
    y_true = np.array(adata.obs[groupby])[samples_bidx]
    y_pred = np.array(circ_adata.obs['scNym'])[samples_bidx]

    n_correct = np.sum(y_true==y_pred)
    n_total = len(y_true)
    acc = n_correct/n_total

    accession_keys_str = [
        '::'.join(x) for x in accession_keys
    ]
    search_df = pd.DataFrame(
        columns = accession_keys_str + ['acc'],
        index=[search_number],
    )
    search_df.loc[search_number] = params + (acc,)
    search_df['test_source_acc'] = s2t_source_test_acc
    
    if groupby_eval is not None:
        # compute the test accuracy in the target domain
        # here, we use the predictions made by the source2target
        # model stored in `adata.obs["scNym"]`.
        samples_bidx = adata.obs[groupby]=='Unlabeled'
        y_true = np.array(adata.obs[groupby_eval])[samples_bidx]
        y_pred = np.array(adata.obs['scNym'])[samples_bidx]
        n_correct = np.sum(y_true == y_pred)
        test_acc = n_correct/len(y_true)
        search_df['test_target_acc'] = 'None'
        search_df.loc[search_number, 'test_target_acc'] = test_acc
    
    search_df.to_csv(osp.join(t2s_out_path, 'result.csv'))

    return search_df


def scnym_tune(
    adata: AnnData,
    groupby: str,
    parameters: dict,
    search: str='grid',
    base_config: str='no_new_identity',
    n_points: int=100,
    out_path: str='./scnym_tune',
    hold_out_only: bool=True,
    groupby_eval: str=None,
    n_splits: int=1,
) -> (pd.DataFrame, dict):
    '''Perform hyperparameter tuning of an scNym model using 
    circular cross-validation.
    
    Parameters
    ----------
    adata : anndata.AnnData
        [Cells, Genes] experiment for optimization.
    groupby : str
        annotation column in `.obs`.
    parameters : dict
        key:List[value] pairs of parameters to use for 
        hyperparameter tuning.
    base_config : str
        one of {"no_new_identity", "new_identity_discovery"}.
        base configuration for model training that described
        default parameters, not explicitly provided in
        `parameters`.
    search : str
        {"grid", "random"} perform either a random or grid
        search over `parameters`.        
    n_points : int
        number of random points to search if `search == "random"`.
    out_path : str
        path for intermediary files during hyperparameter tuning.
    hold_out_only : bool
        evaluate the circular accuracy only on a held-out set of 
        training data, not used in the training of the first
        source -> target model.
    groupby_eval : str
        column in `adata.obs` containing ground truth labels
        for the "Unlabeled" dataset to use for evaluation.
    n_splits : int
        number of train/test/val splits to perform for tuning.
        performs at least 5-fold splitting and uses a subset of
        the folds if `n_splits < 5`.
        
    Returns
    -------
    tuning_results : pd.DataFrame
        [n_points, (parameters,) + (circ_acc, circ_loss)]
    best_parameter_set : dict
        a configuration describing the best parameter set tested.
        
    Examples
    --------
    >>> # `adata` contains labels in `.obs["annotations"]` where
    ... # the target dataset is labeled "Unlabeled"
    >>> tuning_results, best_parameters = scnym_tune(
    ...   adata=adata,
    ...   groupby="annotations",
    ...   parameters={
    ...     "weight_decay": [1e-6, 1e-5, 1e-4],
    ...     "unsup_max_weight": [0.1, 1., 10.],
    ...   },
    ...   base_config="no_new_identity",
    ...   search="grid",
    ...   out_path="./scnym_tuning",
    ...  n_splits=5,
    ... )
        
    Notes
    -----
    Circular/Reverse cross-validation evaluates the impact of hyperparameter
    selection in semi-supervised learning settings using the training data,
    training labels, and target data, but not the target labels.
    
    This is achieved by training a model :math:`f` on the training set, then 
    predicting "pseudolabels" for the target set.
    A second model :math:`g` is then trained on the target data and
    the associated pseudolabels.
    The model :math:`g` is used to predict labels for the *training* set.
    The accuracy of this "reverse" prediction is then used as an estimate
    of the effectiveness of a hyperparameter set.
    '''
    os.makedirs(out_path, exist_ok=True)
    
    # get the base configuration dict
    # configurations have one layer of nested dictionaries within
    config = CONFIGS.get(base_config, None)
    if config is None:
        msg = f'{base_config} is not a valid base configuration.'
        raise ValueError(msg)
        
    #################################################
    # get all possible combinations of parameters
    #################################################
    # `_get_keys_and_list` traverses a nested dictionary and
    # returns a List[list] of sequential keys to access each
    # item in `parameter_ranges`.
    # items in `parameter_ranges: List[list]` are lists of 
    # values for the parameter specified in `accession_keys`.
    accession_keys, parameter_ranges = _get_keys_and_list(
        parameters
    )
    # find all possible combinations of parameters
    # each item in `param_sets` is a tuple of parameter values
    # each element in the tuple matches the keys in `keys` with 
    # the same index.
    param_sets = list(
        itertools.product(
            *parameter_ranges,
        )
    )
    
    #################################################
    # select a set of parameters to search
    #################################################
    if search.lower() == 'random':
        # perform a random search by subsetting grid points
        param_idx = np.random.choice(
            len(param_sets),
            size=n_points,
            replace=False,
        )
    else:
        param_idx = range(len(param_sets))
        
    #################################################
    # set a common train/test/val split for all params
    #################################################
    
    splits_provided = (
        'scNym_split_0' in adata.obs.columns
    )
    splits_provided = (
        splits_provided or 'scNym_split' in adata.obs.columns
    )
    
    if not splits_provided:
        split_data(
            adata,
            groupby=groupby,
            n_splits=n_splits,
        )
    elif n_splits == 1 and 'scNym_split' in adata.obs.columns:
        adata.obs['scNym_split_0'] = adata.obs['scNym_split']
    elif n_splits > 1 and splits_provided:
        # check that we have the relevant split for each fold
        splits_correct = True
        for s in range(n_splits):
            splits_correct = (
                splits_correct & (f'scNym_split_{s}' in adata.obs.columns)
            )
        if not splits_correct:
            msg = '"scNym_split_" was provided with `n_splits>1.\n'
            msg += 'f"scNym_split_{n}"" must be present in `adata.obs` for all {n} in `range(n_splits)`\n'
            raise ValueError(msg)
    else:
        msg = 'invalid argument for n_splits'
        raise ValueError(msg)

    #################################################
    # circular training for each parameter set
    #################################################
    
    accession_keys_str = [
        '::'.join(x) for x in accession_keys
    ]
    
    search_results = []
    search_config_store = []
    for search_number, idx in enumerate(param_idx):
        # get the parameter set
        params = param_sets[idx]
        # update the base config with search parameters
        search_config = copy.deepcopy(config)
        for p_i in range(len(params)):
            keys2update = accession_keys[p_i]
            value2set   = params[p_i]
            # updates in place
            _updated_nested(
                search_config,
                keys2update,
                value2set,
            )
        
        # disable checkpoints, tensorboard to reduce I/O
        search_config['save_freq'] = 10000
        search_config['tensorboard'] = False
        # add search number to config
        search_config['search_number'] = search_number
        
        search_config_store.append(
            copy.deepcopy(search_config),
        )
        print('searching config:')
        pprint.pprint(search_config)
        
        for split_number in range(n_splits):
            # set the relevant split indices
            adata.obs['scNym_split'] = (
                adata.obs[f'scNym_split_{split_number}']
            )
            # set the split number
            split_config = copy.deepcopy(search_config)
            split_config['split_number'] = split_number
            search_df = _circular_train(
                search_config=split_config,
                params=params,
                adata=adata,
                groupby=groupby,
                out_path=out_path,
                accession_keys=accession_keys,
                hold_out_only=hold_out_only,
                groupby_eval=groupby_eval,
            )
            # add the split information
            search_df['split_number'] = split_number
            search_df['search_number'] = search_number
            # save results
            search_results.append(search_df)
        
            print('search results:')
            print(search_df)
    
    # concatenate
    search_results = pd.concat(search_results, 0)
    best_idx = np.argmax(search_results['acc'])
    best_search = int(
        search_results.iloc[best_idx]['search_number']
    )
    
    best_config = search_config_store[best_search]
    print('>>>>>>')
    print('Best config')
    print(best_config)
    print('>>>>>>')
    print()
    return search_results, best_config



'''Test the model Trainer'''
import pytest
import os
import os.path as osp
import sys
import numpy as np
import torch

sys.path.append('../')


def test_trainer():
    import scnym
    import scanpy as sc
    import pandas as pd

    # load 10x human PBMC data as a sample
    adata = sc.datasets.pbmc3k()
    sc.pp.normalize_per_cell(adata, counts_per_cell_after=1e6)
    sc.pp.log1p(adata)
    sc.pp.highly_variable_genes(adata, n_top_genes=2000)
    sc.pp.pca(adata)
    sc.pp.neighbors(adata, n_neighbors=15)
    # generate clusters to use as class labels
    sc.tl.leiden(adata, resolution=0.5, key_added='leiden')
    
    adata.obs['class'] = pd.Categorical(
        adata.obs['leiden'],
    ).codes

    X = adata.X
    y = np.array(adata.obs['class'])
    
    ds = scnym.dataprep.SingleCellDS(
        X=X,
        y=y,
    )
    
    l_dl = torch.utils.data.DataLoader(ds, batch_size=256)
    t_dl = torch.utils.data.DataLoader(ds, batch_size=256)
    u_dl = torch.utils.data.DataLoader(ds, batch_size=256)
    
    dataloaders = {
        'train': l_dl,
        'val'  : t_dl,
    }
    
    criterion = scnym.trainer.cross_entropy
    
    model = scnym.model.CellTypeCLF(
        n_genes=X.shape[1],
        n_cell_types=len(np.unique(y)),
        n_hidden=128,
        n_layers=1,
        init_dropout=0.1,
    )
    if torch.cuda.is_available():
        model = model.cuda()
    
    optimizer = torch.optim.Adadelta(
        model.parameters(),
    )
    trainer_kwgs = {
        'verbose' : True,
        'scheduler' : None,
        'exp_name' : 'tmp',
        'out_path' : './tmp/',
        'model' : model,
        'optimizer' : optimizer,
        'criterion' : criterion,
        'dataloaders': dataloaders,
        'n_epochs': 2,
    }
    
    
    # initialize a standard data loader
    T = scnym.trainer.Trainer(
        **trainer_kwgs
    )
    T.train()
    
    # initialize a semi-supervised data loader
    model = scnym.model.CellTypeCLF(
        n_genes=X.shape[1],
        n_cell_types=len(np.unique(y)),
        n_hidden=128,
        n_layers=1,
    )
    if torch.cuda.is_available():
        model = model.cuda()
    
    optimizer = torch.optim.Adadelta(
        model.parameters(),
    )    
    
    ssl_kwargs = {}
    unsup_criterion = torch.nn.MSELoss(reduction='none')
    USL = scnym.trainer.MixMatchLoss(
        alpha=0.3,
        unsup_criterion=unsup_criterion,
        sup_criterion=trainer_kwgs['criterion'],
        decay_coef=ssl_kwargs.get('decay_coef', 0.997),
        mean_teacher=False,
        augment=scnym.dataprep.AUGMENTATION_SCHEMES['None'],
        n_augmentations=1,
        T=0.5,
        augment_pseudolabels=False,
        pseudolabel_min_confidence=0.6,
    )
    
    weight_schedule = scnym.trainer.ICLWeight(
        ramp_epochs=1,
        max_unsup_weight=0.,
        burn_in_epochs = 0,
        sigmoid = True,
    )
    
    T = scnym.trainer.SemiSupervisedTrainer(
        unsup_dataloader=u_dl,
        unsup_criterion=USL,
        unsup_weight=weight_schedule,
        **trainer_kwgs,
    )
    T.train()
    assert T.best_loss < 0.1
    
    # train with a DANN
    model = scnym.model.CellTypeCLF(
        n_genes=X.shape[1],
        n_cell_types=len(np.unique(y)),
        n_hidden=128,
        n_layers=1,
    )
    if torch.cuda.is_available():
        model = model.cuda()

    optimizer = torch.optim.Adadelta(
        model.parameters(),
    )    
    
    ssl_kwargs = {}
    unsup_criterion = torch.nn.MSELoss(reduction='none')
    USL = scnym.trainer.MixMatchLoss(
        alpha=0.3,
        unsup_criterion=unsup_criterion,
        sup_criterion=trainer_kwgs['criterion'],
        decay_coef=ssl_kwargs.get('decay_coef', 0.997),
        mean_teacher=False,
        augment=scnym.dataprep.AUGMENTATION_SCHEMES['None'],
        n_augmentations=1,
        T=0.5,
        augment_pseudolabels=False,
        pseudolabel_min_confidence=0.6,
    )
    dan_criterion = scnym.trainer.DANLoss(
        model=model,
        dan_criterion=scnym.trainer.cross_entropy,
        use_conf_pseudolabels=True,
    )
    
    weight_schedule = scnym.trainer.ICLWeight(
        ramp_epochs=1,
        max_unsup_weight=0.1,
        burn_in_epochs = 0,
        sigmoid = True,
    )
    dan_weight = scnym.trainer.ICLWeight(
        ramp_epochs=1,
        max_unsup_weight=0.1,
        burn_in_epochs = 0,
        sigmoid = True,
    )    
    
    T = scnym.trainer.SemiSupervisedTrainer(
        unsup_dataloader=u_dl,
        unsup_criterion=USL,
        unsup_weight=weight_schedule,
        dan_criterion=dan_criterion,
        dan_weight=dan_weight,
        **trainer_kwgs,
    )
    T.train()
    assert T.best_loss < 0.1    
    
    return


def main():
    test_trainer()


if __name__ == '__main__':
    main()

    
    
    


import pytest
import os
import os.path as osp
import sys
import numpy as np
import torch

sys.path.append('../')

def test_fit_model_sup():
    '''Test the model fitting function for supervised learning.
    '''
    import scnym
    import scanpy as sc
    torch.manual_seed(1)
    np.random.seed(1)     

    # load 10x human PBMC data as a sample
    adata = sc.datasets.pbmc3k()
    # generate fake class labels
    adata.obs['class'] = np.random.randint(
        0,
        3,
        adata.shape[0],
    )
    
    X = adata.X
    y = np.array(adata.obs['class'])
    
    # test training without unlabeled counts    
    os.makedirs('./tmp', exist_ok=True)
    acc, loss = scnym.main.fit_model(
        X=X,
        y=y,
        traintest_idx=np.arange(X.shape[0]//2),
        val_idx=np.arange(X.shape[0]//2, X.shape[0]),
        batch_size=256,
        n_epochs=1,
        lr=1e-5,
        optimizer_name='adadelta',
        weight_decay=1e-4,
        ModelClass=scnym.model.CellTypeCLF,
        out_path='./tmp',
        n_genes=adata.shape[1],
        mixup_alpha=0.3,
        n_hidden=128,
        n_layers=2,
        residual=False,
    )
    
    return

def test_fit_model_ssl():
    '''Test the model fitting function for semi-supervised
    learning.
    '''
    import scnym
    import scanpy as sc
    torch.manual_seed(1)
    np.random.seed(1)     

    # load 10x human PBMC data as a sample
    adata = sc.datasets.pbmc3k()
    # generate fake class labels
    adata.obs['class'] = np.random.randint(
        0,
        3,
        adata.shape[0],
    )
    
    X = adata.X
    y = np.array(adata.obs['class'])
    # test training with unlabeled counts
    os.makedirs('./tmp', exist_ok=True)    
    acc, loss = scnym.main.fit_model(
        X=X,
        y=y,
        traintest_idx=np.arange(X.shape[0]//2),
        val_idx=np.arange(X.shape[0]//2, X.shape[0]),
        batch_size=256,
        n_epochs=2,
        lr=1e-5,
        optimizer_name='adadelta',
        weight_decay=1e-4,
        ModelClass=scnym.model.CellTypeCLF,
        out_path='./tmp',
        n_genes=adata.shape[1],
        mixup_alpha=0.3,
        n_hidden=128,
        n_layers=2,
        residual=False,
        unlabeled_counts=X[:1000],
        unsup_max_weight=1e-5,
        ssl_method='mixmatch',
        ssl_kwargs={
            'augment': 'None',
            'n_augmentations': 1,
            'augment_pseudolabels': False,
            'T': 1.,
        },
    )
    return


def test_fit_model_ssl_ce():
    '''Test the model fitting function for semi-supervised
    learning with cross-entropy loss.
    '''
    import scnym
    import scanpy as sc
    torch.manual_seed(1)
    np.random.seed(1)     

    # load 10x human PBMC data as a sample
    adata = sc.datasets.pbmc3k()
    # generate fake class labels
    adata.obs['class'] = np.random.randint(
        0,
        3,
        adata.shape[0],
    )
    
    X = adata.X
    y = np.array(adata.obs['class'])
    # test training with unlabeled counts
    os.makedirs('./tmp', exist_ok=True)    
    acc, loss = scnym.main.fit_model(
        X=X,
        y=y,
        traintest_idx=np.arange(X.shape[0]//2),
        val_idx=np.arange(X.shape[0]//2, X.shape[0]),
        batch_size=256,
        n_epochs=2,
        lr=1e-5,
        optimizer_name='adadelta',
        weight_decay=1e-4,
        ModelClass=scnym.model.CellTypeCLF,
        out_path='./tmp',
        n_genes=adata.shape[1],
        mixup_alpha=0.3,
        n_hidden=128,
        n_layers=2,
        residual=False,
        unlabeled_counts=X[:1000],
        unsup_max_weight=1e-5,
        ssl_method='mixmatch',
        ssl_kwargs={
            'augment': 'None',
            'n_augmentations': 1,
            'augment_pseudolabels': False,
            'T': 1.,
            'unsup_criterion': 'crossentropy',
        },
    )
    return


def test_fit_model_ssl_dan():
    '''Test the model fitting function for semi-supervised
    learning with domain adaptation.
    '''
    import scnym
    import scanpy as sc
    torch.manual_seed(1)
    np.random.seed(1)    

    # load 10x human PBMC data as a sample
    adata = sc.datasets.pbmc3k()
    # generate fake class labels
    adata.obs['class'] = np.random.randint(
        0,
        3,
        adata.shape[0],
    )
    
    X = adata.X
    y = np.array(adata.obs['class'])
    # test training with unlabeled counts    
    acc, loss = scnym.main.fit_model(
        X=X,
        y=y,
        traintest_idx=np.arange(X.shape[0]//2),
        val_idx=np.arange(X.shape[0]//2, X.shape[0]),
        batch_size=256,
        n_epochs=2,
        lr=1e-5,
        optimizer_name='adadelta',
        weight_decay=1e-4,
        ModelClass=scnym.model.CellTypeCLF,
        out_path='./tmp',
        n_genes=adata.shape[1],
        mixup_alpha=0.3,
        n_hidden=128,
        n_layers=2,
        residual=False,
        unlabeled_counts=X[:1000],
        unsup_max_weight=1e-5,
        ssl_method='mixmatch',
        ssl_kwargs={
            'augment': 'None',
            'n_augmentations': 1,
            'augment_pseudolabels': False,
            'T': 1.,
            'unsup_criterion': 'mse',
            'dan_criterion': True,
            'dan_ramp_epochs': 100,
            'dan_burn_in_epochs': 0,
            'dan_max_weight': 1.,
        },
    )
    return

def test_fit_model_dan_multdom():
    '''Test the model fitting function for semi-supervised
    learning with domain adaptation across >2 domains.
    '''
    import scnym
    import scanpy as sc
    torch.manual_seed(1)
    np.random.seed(1)    

    # load 10x human PBMC data as a sample
    adata = sc.datasets.pbmc3k()
    # generate fake class labels
    adata.obs['class'] = np.random.randint(
        0,
        3,
        adata.shape[0],
    )
    
    X = adata.X
    y = np.array(adata.obs['class'])
    
    input_domain = np.zeros(X.shape[0])
    unlabeled_domain = np.ones(1000)
    unlabeled_domain[500:] += 1
    
    n_domains = int(np.max([
        input_domain.max(), unlabeled_domain.max()
    ])) + 1
    print('Max domain: ')
    # test training with unlabeled counts    
    acc, loss = scnym.main.fit_model(
        X=X,
        y=y,
        traintest_idx=np.arange(X.shape[0]//2),
        val_idx=np.arange(X.shape[0]//2, X.shape[0]),
        batch_size=256,
        n_epochs=2,
        lr=1e-5,
        optimizer_name='adadelta',
        weight_decay=1e-4,
        ModelClass=scnym.model.CellTypeCLF,
        out_path='./tmp',
        n_genes=adata.shape[1],
        mixup_alpha=0.3,
        n_hidden=128,
        n_layers=2,
        residual=False,
        unlabeled_counts=X[:1000],
        unsup_max_weight=1e-5,
        ssl_method='mixmatch',
        ssl_kwargs={
            'augment': 'None',
            'n_augmentations': 1,
            'augment_pseudolabels': False,
            'T': 1.,
            'unsup_criterion': 'mse',
            'dan_criterion': True,
            'dan_ramp_epochs': 100,
            'dan_burn_in_epochs': 0,
            'dan_max_weight': 1.,
        },
        input_domain=input_domain,
        unlabeled_domain=unlabeled_domain,
    )
    return


def main():
    test_fit_model_dan_multdom()
    return


if __name__ == '__main__':
    main()



import pytest
import os
import os.path as osp
import sys
import numpy as np
import torch

sys.path.append('../')


def test_model_build():
    '''Test construction of a standard model'''
    import scnym
    torch.manual_seed(1)
    np.random.seed(1)    
    
    model = scnym.model.CellTypeCLF(
        n_genes=10000,
        n_cell_types=10,
        n_hidden=64,
        n_layers=2,
        residual=False,
        init_dropout=0.0,
    )
    
    assert hasattr(model, 'classif')
    assert ~hasattr(model, 'dan')
    
    
    model = scnym.model.CellTypeCLF(
        n_genes=10000,
        n_cell_types=10,
        n_hidden=64,
        n_layers=2,
        residual=False,
        init_dropout=0.0,
        track_running_stats=False,
    )
    
    return


def test_model_dan_build():
    '''Test construction of a DANN model'''
    import scnym
    torch.manual_seed(1)
    np.random.seed(1)    
    
    model = scnym.model.CellTypeCLF(
        n_genes=10000,
        n_cell_types=10,
        n_hidden=64,
        n_layers=2,
        residual=False,
        init_dropout=0.0,
    )
    
    assert hasattr(model, 'classif')
    
    dann = scnym.model.DANN(
        model=model,
        n_domains=2,
    )
        
    assert hasattr(dann, 'embed')
    assert hasattr(dann, 'domain_clf')
    
    return


def test_model_save_load():
    '''Test saving and loading of different model types'''
    import scnym
    torch.manual_seed(1)
    np.random.seed(1)    
    
    std_model = scnym.model.CellTypeCLF(
        n_genes=10000,
        n_cell_types=10,        
        n_hidden=64,
        n_layers=2,
        residual=False,
        init_dropout=0.0,
    )
    
    os.makedirs('./tmp', exist_ok=True)
    # save
    torch.save(
        std_model.state_dict(),
        './tmp/std_model.pkl',
    )
    
    # load
    std_model.load_state_dict(
        torch.load(
            './tmp/std_model.pkl',
        )
    )
    
    return


def main():
    test_model_dan_build()
    return


if __name__ == '__main__':
    main()


'''Tools for interpreting trained scNym models'''
import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
import typing
import copy
import warnings


class Salience(object):
    '''
    Performs backpropogation to compute gradients on a target
    class with regards to an input.

    Notes
    -----
    Saliency analysis computes a gradient on a target class
    score :math:`f_i(x)` with regards to some input :math:`x`.


    .. math::

        S_i = \frac{\partial f_i(x)}{\partial x}
    '''

    def __init__(
        self,
        model: nn.Module,
        class_names: np.ndarray,
        gene_names: np.ndarray = None,
        layer_to_hook: int=None,
        verbose: bool=False,
    ) -> None:
        '''
        Performs backpropogation to compute gradients on a target
        class with regards to an input.

        Parameters
        ----------
        model : torch.nn.Module
            trained scNym model.
        class_names : np.ndarray
            list of str names matching output nodes in `model`.
        gene_names : np.ndarray, optional
            gene names for the model.
        layer_to_hook : int
            index of the layer from which to record gradients.
            defaults to the gene level input features.

        Returns
        -------
        None.
        '''
        # ensure class names are unique for each output node
        if len(np.unique(class_names)) != len(class_names):
            msg = '`class_names` must all be unique.'
            raise ValueError(msg)

        self.class_names = np.array(class_names)
        self.n_classes = len(class_names)
        self.verbose = verbose

        # load model into CUDA compute if available
        if torch.cuda.is_available():
            self.model = model.cuda()
        else:
            self.model = model
        # ensure we're not in training mode
        self.model = self.model.eval()

        self.gene_names = gene_names
        
        if layer_to_hook is None:
            self._hook_first_layer_gradients()
        else:
            self._hook_nth_layer_gradients(n=layer_to_hook)
        return

    def _hook_first_layer_gradients(self):
        '''Set up hooks to record gradients from the first linear
        layer into a target tensor.

        References
        ----------
        https://pytorch.org/docs/stable/nn.html#torch.nn.Module.register_backward_hook
        '''

        def _record_gradients(module, grad_in, grad_out):
            '''Record gradients of a layer with the correct input
            shape'''
            self.gradients = grad_in[1]
            if self.verbose:
                print([x.size() if x is not None else 'None' for x in grad_in])
                print('Hooked gradients to: ', module)

        for module in self.model.modules():
            if isinstance(module, nn.Linear) and module.in_features == len(self.gene_names):
                    module.register_backward_hook(_record_gradients)
        return
    
    def _hook_nth_layer_gradients(self, n: int):
        '''Set up hooks to record gradients from an arbitrary layer.

        References
        ----------
        https://pytorch.org/docs/stable/nn.html#torch.nn.Module.register_backward_hook
        '''

        def _record_gradients(module, grad_in, grad_out):
            '''Record gradients of a layer with the correct input
            shape'''
            self.gradients = grad_in[1]
            if self.verbose:
                print([x.size() if x is not None else 'None' for x in grad_in])
                print('Hooked gradients to: ', module)

        module = list(self.model.modules())[n]
        module.register_backward_hook(_record_gradients)
        return    

    def _guided_backprop_hooks(self):
        '''Set up forward and backward hook functions to perform
        "Guided backpropogation"

        Notes
        -----
        Guided backpropogation only passes positive gradients upward through the network.

        Normal backprop:

        .. math::

            f_i^{(l + 1)} = ReLU(f_i^{(l)})

            R_i^{(l)} = (f_i^{(l)} > 0) \cdot R_i^{(l+1)}

        where

        .. math::

            R_i^{(l + 1)} = \frac{\partial f_{out}}{\partial f_i^{l + 1}}


        By contrast, guided backpropogation only passes gradient values where both
        the activates :math:`f_i^{(l)}` and the gradients :math:`R_i^{(l + 1)}` are
        greater than :math:`0`.


        References
        ----------
        https://arxiv.org/pdf/1412.6806.pdf

        https://pytorch.org/docs/stable/nn.html#torch.nn.Module.register_forward_hook        
        https://pytorch.org/docs/stable/nn.html#torch.nn.Module.register_backward_hook
        '''
        def _record_relu_outputs(module, in_, out_):
            '''Store the outputs to each ReLU layer'''
            self.rectified_outputs.append(out_,)
            self.store_rectified_outputs.append(out_,)

        def _clamp_grad(module, grad_in, grad_out):
            '''Clamp ReLU gradients to [0, inf] and return a 
            new gradient to be used in subsequent outputs.
            '''
            self.store_grad.append(grad_in[0])

            grad = grad_in[0].clamp(min=0.0)
            self.store_clamped_grad.append(grad)

            # here we pop the outputs off to ensure that the
            # final output is always the current ReLU layer
            # we're investigating
            last_relu_output = self.rectified_outputs.pop()
            last_relu_output = copy.copy(
                last_relu_output
            )
            last_relu_output[last_relu_output>0] = 1
            rectified_grad = last_relu_output * grad
            
            self.store_rectified_grad.append(rectified_grad)
            return (rectified_grad,)

        self.store_rectified_outputs = []
        self.store_grad = []
        self.store_clamped_grad = []

        for _, module in self.model.named_modules():
            if isinstance(module, nn.ReLU):
                module.register_forward_hook(_record_relu_outputs)
                module.register_backward_hook(_clamp_grad)

        return

    def get_saliency(
        self,
        x: torch.FloatTensor,
        target_class: str,
        guide_backprop: bool = False,
    ) -> torch.FloatTensor:
        '''Compute the saliency of a target class on an input
        vector `x`.

        Parameters
        ----------
        x : torch.FloatTensor
            [1, Genes] vector of gene expression.
        target_class : str
            class in `.class_names` for which to compute gradients.
        guide_backprop : bool
            perform "guided backpropogation" by clamping gradients
            to only positive values at each ReLU.
            see: https://arxiv.org/pdf/1412.6806.pdf

        Returns
        -------
        salience : torch.FloatTensor
            gradients on `target_class` with respect to `x`.
        '''
        if target_class not in self.class_names:
            msg = f'{target_class} is not in `.class_names`'
            raise ValueError(msg)

        target_idx = np.where(
            target_class == self.class_names
        )[0].astype(np.int)
        target_idx = int(target_idx)

        self.model.zero_grad()

        if guide_backprop:
            self.rectified_outputs = []
            self.store_rectified_grad = []            
            self._guided_backprop_hooks()

        # store gradients on the input
        if torch.cuda.is_available():
            x = x.cuda()
        x.requires_grad = True

        # module hook will record gradients here
        self.gradients = torch.zeros_like(x)

        # forward pass
        output = self.model(x)

        # create a [N, C] tensor to store gradients
        target = torch.zeros_like(output)
        # set the target class to `1`, creating a one-hot
        # of the target class
        target[:, target_idx] = 1

        # compute gradients with backprop
        output.backward(
            gradient=target,
        )

        # detach from the graph and move to main memory
        target = target.detach().cpu()

        return self.gradients


    def rank_genes_by_saliency(
        self,
        **kwargs,
    ) -> np.ndarray:
        '''
        Rank genes by saliency for a target class and input.

        Passes **kwargs to `.get_saliency` and uses the output
        to rank genes.

        Returns
        -------
        ranked_genes : np.ndarray
            gene names with high saliency, ranked highest to
            lowest.
        '''
        s = self.get_saliency(**kwargs)
        sort_idx = torch.argsort(s)
        idx = sort_idx[0].numpy()[::-1]
        return self.gene_names[idx.astype(np.int)]

    
class IntegratedGradient(object):
    
    
    def __init__(
        self,
        model: nn.Module,
        class_names: typing.Union[list, np.ndarray],
        gene_names: typing.Union[list, np.ndarray] = None,
        verbose: bool = False,
    ) -> None:
        '''Performs integrated gradient computations for feature attribution
        in scNym models.
        
        Parameters
        ----------
        model : torch.nn.Module
            trained scNym model.
        class_names : list or np.ndarray
            list of str names matching output nodes in `model`.
        gene_names : list or np.ndarray, optional
            gene names for the model.
        verbose : bool
            verbose outputs for stdout.
        
        Returns
        -------
        None.
        
        Notes
        -----
        Integrated gradients are computed as the path integral between a "baseline"
        gene expression vector (all 0 counts) and an observed gene expression vector.
        The path integral is computed along a straight line in the feature space.
        
        Stated formally, we define a our baseline gene expression vector as :math:`x`,
        our observed vector as :math:`x'`, an scnym model :math:`f(\cdot)`, and a 
        number of steps :math:`M` for approximating the integral by Reimann sums.
        
        The integrated gradient :math:`\int \nabla` for a feature :math:`x_i` is then
        
        .. math::
        
            r = \sum_{m=1}^M \partial f(x' + \frac{m}{M}(x - x')) / \partial x_i \\
            \int \nabla_i = (x_i' - x_i) \frac{1}{M} r
        '''
        self.model = copy.deepcopy(model)
        if torch.cuda.is_available():
            self.model = self.model.cuda()
            print('Model loaded on CUDA compute device.')
        self.model.zero_grad()
        for param in self.model.parameters():
            param.requires_grad = False        
        
        self.class_names = class_names
        self.gene_names = gene_names
        self.verbose = verbose
        
        if type(self.class_names) == np.ndarray:
            self.class_names = self.class_names.tolist()
                
        return
    
    
    def get_grad(
        self,
        x: torch.Tensor,
        target_class: str,
    ) -> (torch.Tensor, torch.Tensor):
        '''Get the gradient for a minibatch of observations with respect
        to a target class.
        
        Parameters
        ----------
        x : torch.Tensor
            [Batch, Features] input tensor.
        target_class : str
            target class for gradient computation.
            
        Returns
        -------
        grad : torch.Tensor
            [Batch, Features] feature gradients with respect to the
            target class.
        target : torch.Tensor
            [Batch,] value of the target class score.
        '''
        target_idx = self.class_names.index(target_class)
        
        # store gradients on the input
        if torch.cuda.is_available():
            x = x.cuda()
        x.requires_grad = True
    
        # forward pass through the model
        output = self.model(x)
        sm_output = F.softmax(output, dim=-1)
        
        # get the softmax output on the target class for each
        # observation as a loss
        index = torch.ones(output.size(0)).view(-1, 1) * target_idx
        index = index.long()
        index = index.to(device=sm_output.device)
        # `.gather(dim, index)` takes a dimension number and a tensor
        # of indices size [Batch,] where each val is an integer index
        # grabs the specific element for each observation along the given dim.
        target = sm_output.gather(1, index)
        
        # zero any existing gradients
        self.model.zero_grad()
        if x.grad is not None:
            x.grad.zero_()
        target.backward()
        
        grad = x.grad.detach().cpu()
        
        return grad, target
    
    
    def _check_integration(
        self,
        integrated_grad: torch.Tensor,
    ) -> bool:
        '''Check that the approximation of the path integral is appropriate.
        If we used a sufficient number of steps in the Reimann sum, we should
        find that the gradient sum is roughly equivalent to the difference in
        class scores for the baseline vector and target vector.
        '''
        score_difference = self.raw_scores[-1] - self.raw_scores[0]
        check = torch.isclose(
            integrated_grad.sum(),
            score_difference,            
            rtol=0.1,
        )
        if not check:
            msg = 'integrated gradient magnitude does not match the difference in scores.\n'
            msg += f'magnitude {integrated_grad.sum().item()} vs. {score_difference.item()}.\n'
            msg += 'consider using more steps to estimate the path integral.'
            warnings.warn(msg)
        return check
    
    def get_integrated_gradient(
        self,
        x: torch.Tensor,
        target_class: str,
        M: int=300,
        baseline: torch.Tensor=None,
    ) -> torch.Tensor:
        '''Compute the integrated gradient for a single observation.
        
        Parameters
        ----------
        x : torch.Tensor
            [Features,] input tensor.
        target_class : str
            class in `self.class_names` for optimization.
        M : int
            number of gradient steps to use when approximating
            the path integral.
        baseline : torch.Tensor
            [Features,] baseline gene expression vector to use.
            if `None`, uses the `0` vector.
        
        Returns
        -------
        integrated_grad : torch.Tensor
            [Features,] integrated gradient tensor.
            
        Notes
        -----
        1. Define a difference between the baseline input and observation.
        2. Approximate a linear path between the baseline and observation 
        with `M` steps.
        3. Compute the gradient at each step in the path.
        4. Sum gradients across steps and divide by number of steps.
        5. Elementwise multiply with input features as in saliency.
        '''
        if baseline is None:
            if self.verbose:
                print('Using the 0-vector as a baseline.')
            base = self.baseline_input = torch.zeros(
                (1, len(self.gene_names))
            ).float()
        else:
            base = self.baseline_input = baseline
            if base.dim() > 1 and base.size(0) != 1:
                msg = 'baseline must be a single gene expression vector'
                raise ValueError(msg)
            base = base.view(1, -1)
        
        self.target_class = target_class

        if x.dim() > 1 and x.size(0) == 1:
            # tensor has an empty batch dimension, flatten it
            x = x.view(-1)
        
        # create a batch of observations where each observation is
        # a single step along the path integral
        path = base.repeat((M, 1))
        
        # create a tensor marking the "step number" for each observation
        step = ( (x - base) / M).view(1, -1)
        step_coord = torch.arange(1, M+1).view(-1, 1).repeat((1, path.size(1)))
        
        # add the correct number of steps to fill the path tensor
        path += (step * step_coord)
        
        if self.verbose:
            print('baseline', base.size())
            print(base.sort())
            print('observation', x.size())
            print(x.sort())
            print()
            print('step : ', step.size())
            print(step)
            print('step_coord : ', step_coord.size())
            print(step_coord)
            print('path : ', path.size())
            print(path[0].sort())
            print('-'*3)
            print(path[-1].sort())
                
        # compute the gradient on the input at each step
        # along the path
        gradients = torch.zeros_like(path)
        scores = torch.zeros(path.size(0))
    
        for m in range(M):
            gradients[m, :], target_scores = self.get_grad(
                path[m, :].view(1, -1),
                self.target_class,
            )
            scores[m] = target_scores

        self.raw_gradients = gradients
        self.raw_scores = scores
        self.path = path
        
        # sum gradients and normalize by step number
        integrated_grad = (x - base) * (gradients.sum(0) / M)
        
        self._check_integration(integrated_grad)
        
        return integrated_grad
    
    
class Tessaract(IntegratedGradient):
    
    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        """Tessaract finds a path from a source vector in feature
        space to a destination vector that maximizes the likelihood
        of observing each intermediate position using a trained
        classification model.
        """


'''Test domain adaptation tools'''
import pytest
import os
import os.path as osp
import sys
import numpy as np
import torch

sys.path.append('../')


def test_dan_pseudoconf():
    '''Test pseudolabel confidence mechanism for DAN'''
    import scnym
    import copy
    torch.manual_seed(1)
    np.random.seed(1)    
    
    model = scnym.model.CellTypeCLF(
        n_genes=3,
        n_cell_types=2,
        residual=False,
        init_dropout=0.,
        n_layers=1,
        n_hidden=16,
    )
    
    # create fake data from cell type A
    A = torch.randn(size=(32, 3))
    # create fake data from cell type B
    B = torch.randn(size=(32, 3)) + 5
    
    # create datasets and loaders
    # label cell types A/B with [0, 1]
    X_0 = torch.cat([A[:16], B[:16]], dim=0)
    y_0 = torch.zeros(32)
    y_0[16:] = 1
    
    
    # shift X_1 into a new domain
    X_1 = torch.cat([A[16:], B[16:]], dim=0)
    y_1 = torch.zeros(32)
    y_1[16:] = 1
    
    X_1 = X_1 + 10.
    
    # pack into samples
    labeled_sample = {
        'input': X_0,
        'output': y_0.long(),
    }
    unlabeled_sample = {
        'input': X_1,
        'output': y_1.long(),
    }
    
    print('Fake data:')
    print(labeled_sample['input'][:5])
    print(unlabeled_sample['input'][:5])
    
    
    # setup the DAN loss
    dan = scnym.trainer.DANLoss(
        model=model,
        dan_criterion=scnym.trainer.cross_entropy,
        use_conf_pseudolabels=True,
    )
    
    # get pseudolabel confidence
    pseudolabel_min_confidence = 0.9    
    
    model.eval()
    u_out = model(unlabeled_sample['input'])
    u_sm  = torch.nn.functional.softmax(u_out, dim=1)
    u_hc, u_pred = torch.max(u_sm, dim=1)
    
    pseudolabel_confidence = u_hc > pseudolabel_min_confidence
    print(f'# of conf pseudolabels before training: {torch.sum(pseudolabel_confidence)}')
    
    # compute dan loss
    l = dan(
        labeled_sample=labeled_sample,
        unlabeled_sample=unlabeled_sample,
        weight=0.,
        pseudolabel_confidence=pseudolabel_confidence,
    )
    init_l = l.detach().item()
    # how many examples had confident pseudolabels?
    init_n_conf = dan.n_conf_pseudolabels
    
    # we train the model embedding for several iterations
    # to ensure that the pseudolabel confidence increases
    # over time and that this is reflected in the DANN
    optimizer = torch.optim.AdamW(
        [
            {'params':model.parameters()},
        ],
        weight_decay=1e-4,
    )
    
    print('Optimizing model to increase pseudolabel conf...')
    ce_loss = torch.nn.CrossEntropyLoss()
    losses = []
    model.train()
    for i in range(50):
        optimizer.zero_grad()
        
        x_out = model(labeled_sample['input'])
        l = ce_loss(x_out, labeled_sample['output'])
        l.backward()
        
        optimizer.step()
        
        losses.append(l.detach().cpu().item())
    print('Supervised classifier losses:')
    print(f'Initial: {losses[0]}')
    print(f'Final  : {losses[-1]}')
    assert losses[-1] < losses[0]
    
    # check that the pseudolabel confidence is now increased
    model.eval()
    u_out = model(unlabeled_sample['input'])
    u_sm  = torch.nn.functional.softmax(u_out, dim=1)
    u_hc, u_pred = torch.max(u_sm, dim=1)    
    pseudolabel_confidence = u_hc > pseudolabel_min_confidence
    
    l = dan(
        labeled_sample=labeled_sample,
        unlabeled_sample=unlabeled_sample,
        weight=0.,
        pseudolabel_confidence=pseudolabel_confidence,
    )
    final_l = l.detach().item()
    # how many examples had confident pseudolabels?
    final_n_conf = dan.n_conf_pseudolabels
    
    print(f'Initial conf pseudolabels: {init_n_conf}')
    print(f'Final   conf pseudolabels: {final_n_conf}')
    assert final_n_conf > init_n_conf
    
    ######################################################
    # test training the DAN with pseudolabel confidence
    ######################################################
    
    # reset the model parameters
    model = scnym.model.CellTypeCLF(
        n_genes=3,
        n_cell_types=2,
        residual=False,
        init_dropout=0.,
        n_layers=1,
        n_hidden=16,
    )
    # setup a teacher model
    teacher = copy.deepcopy(model)
    teacher.eval()    
    
    dan = scnym.trainer.DANLoss(
        model=model,
        dan_criterion=scnym.trainer.cross_entropy,
        use_conf_pseudolabels=True,
    )
    
    # train the DAN and the model concurrently, but do not pass
    # DAN weigths to the classifier
    # both classification and DAN losses should *decrease*
    print('Optimizing model and DAN...')
    ce_losses = []
    dan_losses = []
    
    optimizer = torch.optim.AdamW(
        [
            {'params':model.parameters()},
            {'params':dan.dann.domain_clf.parameters()},
        ],
        weight_decay=1e-4,
    )
    
    pseudolabel_min_confidence = 0.8
    for i in range(200):
        optimizer.zero_grad()
        
        x_out = model(labeled_sample['input'])
        ce_l = ce_loss(x_out, labeled_sample['output'])
        
        # generate pseudolabels
        teacher.load_state_dict(model.state_dict())
        teacher.eval()
        u_out = teacher(unlabeled_sample['input'])
        u_sm  = torch.nn.functional.softmax(u_out, dim=1)
        u_hc, u_pred = torch.max(u_sm, dim=1)
        pseudolabel_confidence = u_hc > pseudolabel_min_confidence    
        
        dan_l = dan(
            labeled_sample=labeled_sample,
            unlabeled_sample=unlabeled_sample,
            weight=0.,
            pseudolabel_confidence=pseudolabel_confidence,
        )
        
        l = ce_l + dan_l
        l.backward()
        
        ce_losses.append(ce_l.detach().cpu().item())
        dan_losses.append(dan_l.detach().cpu().item())        
        
        # check that gradients are flowing to the model
        model_grad_mag = 0.
        for p in model.classif.parameters():
            if p.grad is not None:
                model_grad_mag += (p.grad**2).sum()
        assert model_grad_mag > 0.
        
        # check that gradients flow to the DANN domain classifier
        dan_grad_mag = 0.
        for p in dan.dann.domain_clf.parameters():
            if p.grad is not None:
                dan_grad_mag += (p.grad**2).sum()
        assert dan_grad_mag > 0.
        if i == 0 or i == 49:
            print(f'Model grad mag: {model_grad_mag}')
            print(f'DANN  grad mag: {dan_grad_mag}')
        
        optimizer.step()
        
        losses.append(l.detach().cpu().item())
        
    # check that losses decreased
    print(f'Initial CE  loss: {ce_losses[0]}')
    print(f'Final   CE  loss: {ce_losses[-1]}')

    print(f'Initial DAN loss: {dan_losses[0]}')
    print(f'Final   DAN loss: {dan_losses[-1]}')
    
    assert ce_losses[-1] < ce_losses[0]
    assert dan_losses[-1] < dan_losses[0]
    
    return


def test_dan_train():
    '''Test training a DANN on fake data that is easy
    to discriminate from three fake domains'''
    import scnym
    torch.manual_seed(1)
    np.random.seed(1)    
        
    model = scnym.model.CellTypeCLF(
        n_genes=3,
        n_cell_types=2,
        residual=False,
        init_dropout=0.,
        n_layers=1,
        n_hidden=16,
    )
    
    # create fake data from cell type A
    A = torch.randn(size=(48, 3))
    # create fake data from cell type B
    B = torch.randn(size=(48, 3)) + 5
    
    # setup the DAN loss
    dan = scnym.trainer.DANLoss(
        model=model,
        dan_criterion=scnym.trainer.cross_entropy,
        n_domains = 3,
    )
    
    # create datasets and loaders
    # label cell types A/B with [0, 1]
    X_0 = torch.cat([A[:16], B[:16]], dim=0)
    y_0 = torch.zeros(32)
    y_0[16:] = 1
    
    
    # shift X_1 into a new domain 1
    X_1 = torch.cat([A[16:32], B[16:32]], dim=0)
    y_1 = torch.zeros(32)
    y_1[16:] = 1
    
    X_1 = X_1 + 1000.
    
    # shift X_2 into new domain 2
    X_2 = torch.cat([A[32:], B[32:]], dim=0)
    y_2 = torch.zeros(32)
    y_2[16:] = 1
    
    X_2 = X_2 - 1000.
    
    # pack into samples
    labeled_domain = torch.cat(
        [torch.zeros(32), torch.ones(32)],
    ).long()
    labeled_domain = torch.nn.functional.one_hot(
        labeled_domain, 
        num_classes=3,
    )
    labeled_sample = {
        'input': torch.cat([X_0, X_2], 0),
        'output': torch.cat([y_0, y_2], 0),
        'domain': labeled_domain,
    }
    unlabeled_domain = (torch.zeros(32) + 2).long()
    unlabeled_domain = torch.nn.functional.one_hot(unlabeled_domain, num_classes=3)
    unlabeled_sample = {
        'input': X_1,
        'output': y_1,
        'domain': unlabeled_domain,
    }
    
    print('Fake data:')
    print(labeled_sample['input'][:5])
    print(unlabeled_sample['input'][:5])
    print('Fake domains:')
    print(labeled_sample['domain'][:5])
    print(unlabeled_sample['domain'][:5])    
    
    # compute dan loss
    l = dan(
        labeled_sample=labeled_sample,
        unlabeled_sample=unlabeled_sample,
        weight=0.,
    )
    init_l = l.detach().item()
    
    # run a few backprop iterations to make sure
    # we can decrease the loss and train the DANN
    optimizer = torch.optim.AdamW(
        [
            {'params':model.parameters()},
            {'params': dan.dann.domain_clf.parameters()},
        ],
        weight_decay=1e-4,
    )
    
    # we scale the reverse gradients passed to the embedding down
    # to 0, so the loss should decrease as we train
    print('Optimizing...')    
    losses = []    
    for i in range(50):
        optimizer.zero_grad()
        l = dan(
            labeled_sample=labeled_sample,
            unlabeled_sample=unlabeled_sample,
            weight=0.,
        )
        l.backward()
        
        # check that gradients are being cut off before
        # flowing back to the model
        model_grad_mag = 0.
        for p in model.classif.parameters():
            if p.grad is not None:
                model_grad_mag += (p.grad**2).sum()
        assert model_grad_mag == 0.
        
        # check that gradients flow to the DANN domain classifier
        dan_grad_mag = 0.
        for p in dan.dann.domain_clf.parameters():
            if p.grad is not None:
                dan_grad_mag += (p.grad**2).sum()
        assert dan_grad_mag > 0.
        if i == 0 or i == 49:
            print(f'Model grad mag: {model_grad_mag}')
            print(f'DANN  grad mag: {dan_grad_mag}')
        
        optimizer.step()
        
        losses.append(l.detach().cpu().item())
        
    final_l = l.detach().item()
    
    if not (init_l > final_l):
        msg = f'Losses did not decrease for domain clf train:\n{losses}'
        raise ValueError(msg)
    else:
        print('Decreasing DA losses by domain clf train')
        print(losses)
        
    # now make sure we are increasing the losses when we only train the
    # embedding model, rather than the DANN itself.
    optimizer = torch.optim.AdamW(
        [
            {'params':model.parameters()},
        ],
        weight_decay=1e-4,
    )
    losses = []

    # get a copy of the DANN parameters to show that 
    # they aren't changing as we optimize the model embedding
    dann_p_t0 = list(dan.dann.domain_clf.parameters())[0]
    for i in range(50):
        optimizer.zero_grad()
        l = dan(
            labeled_sample=labeled_sample,
            unlabeled_sample=unlabeled_sample,
            weight=1.,
        )
        l.backward()
        optimizer.step()
        
        # check that gradients are flowing back to the model
        model_grad_mag = 0.
        for p in model.classif.parameters():
            if p.grad is not None:
                model_grad_mag += (p.grad**2).sum()
        assert model_grad_mag > 0.
        
        # check that gradients are going to the DANN
        # classifier, even if we're not updating them
        dan_grad_mag = 0.
        for p in dan.dann.domain_clf.parameters():
            if p.grad is not None:
                dan_grad_mag += (p.grad**2).sum()
        assert dan_grad_mag > 0.
        if i == 0 or i == 49:
            print(f'Model grad mag: {model_grad_mag}')
            print(f'DANN  grad mag: {dan_grad_mag}')
        
        # check that DANN parameters aren't changing
        dann_p_t1 = list(dan.dann.domain_clf.parameters())[0]
        assert torch.all(dann_p_t0 == dann_p_t1)
        dann_p_t0 = dann_p_t1
        
        losses.append(l.detach().cpu().item())  
        
    if not (losses[0] < losses[-1]):
        msg = f'Losses did not increase for +ive loss:\n{losses}'
        raise ValueError(msg)
    else:
        print('Increasing DA losses with +ive term')
        print(losses)        
        
    return


if __name__ == '__main__':
    #test_dan_pseudoconf()
    #print()
    test_dan_train()


import pytest
import os
import os.path as osp
import sys
import numpy as np
import torch

sys.path.append('../')


def test_samplemixup():
    '''Confirms that MixUp is properly saving the
    sample identity of origin for MixMatch.
    '''
    import scnym
    torch.manual_seed(1)
    np.random.seed(1)    
    
    mixup = scnym.dataprep.SampleMixUp(
        alpha=0.3,
        keep_dominant_obs=True,
    )
    
    # create two tensors with two boolean
    # features. One tensor is positive for
    # feature 0, the other for feature 1
    n = 3
    A = torch.zeros((n, 2))
    A[:, 0] = 1
    B = torch.zeros((n, 2))
    B[:, 1] = 1
    
    # create a dummy sample of these
    # two tensors
    C = torch.cat([A, B], dim=0)
    
    sample = {
        'input' : C,
        'output': C,
        'some_other_tensor': C,
    }
    
    # mix the dummy sample
    sample_mixed = mixup(
        sample,
    )
    
    # if our MixUp worked properly, the first `n`
    # examples should be dominated by tensor `A`,
    # and the latter `n` by tensor `B`.
    _, dom_idx = torch.max(sample_mixed['input'], dim=1)
    assert torch.all(dom_idx[:n] == 0)
    assert torch.all(dom_idx[n:] == 1)
    
    # double check that the remaining tensor
    # was also mixed properly
    _, dom_idx = torch.max(sample_mixed['some_other_tensor'], dim=1)
    assert torch.all(dom_idx[:n] == 0), 'other tensor was not mixed'
    assert torch.all(dom_idx[n:] == 1), 'other tensor was not mixed'
    return


def test_mixmatch_forward():
    '''Perform a forward pass through the MixMatch objective
    using a dummy model and resampling augmentations
    
    Notes
    -----
    Does not verify that computations are correct, merely that
    the forward pass completes without error.
    '''
    
    import scanpy as sc
    import scnym
    from scnym.dataprep import SingleCellDS
    torch.manual_seed(1)
    np.random.seed(1)    
    
    # load 10x human PBMC data as a sample
    adata = sc.datasets.pbmc3k()
    sc.pp.filter_cells(adata, min_counts=100)
    sc.pp.filter_genes(adata, min_cells=100)
    sc.pp.normalize_per_cell(adata, counts_per_cell_after=1e6)
    sc.pp.log1p(adata)
    
    # generate fake class labels
    adata.obs['class'] = np.random.randint(
        0,
        3,
        adata.shape[0],
    )
    
    # create dataloaders
    
    d_ds = SingleCellDS(
        X = adata.X.toarray(),
        y = np.array(adata.obs['class']),
    )
    u_ds = SingleCellDS(
        X = adata.X.toarray(),
        y = np.zeros(adata.shape[0]),
    )
    
    d_dl = torch.utils.data.DataLoader(d_ds, batch_size=16)
    u_dl = torch.utils.data.DataLoader(u_ds, batch_size=16)
    
    # draw a batch from each
    d_batch = next(iter(d_dl))
    u_batch = next(iter(u_dl))
    
    # initialize mixmatch using a dummy model
    unsup_criterion = torch.nn.MSELoss(reduction='none')
    sup_criterion = scnym.trainer.cross_entropy
    
    augment = scnym.dataprep.AUGMENTATION_SCHEMES['log1p_drop']
    
    model = scnym.model.CellTypeCLF(
        n_layers=1,
        n_hidden=128,
        residual=False,
        init_dropout=0.0,
        n_genes=adata.shape[1],
        n_cell_types=3,
    )
    
    K = 2
    mixmatch = scnym.trainer.MixMatchLoss(
        unsup_criterion=unsup_criterion,
        sup_criterion=sup_criterion,
        augment=augment,
        alpha=0.3,
        n_augmentations=K,
    )

    # mixmatch forward pass
    loss_sup, loss_unsup, sup_outputs = mixmatch(
        model,
        unlabeled_sample=u_batch, 
        labeled_sample=d_batch,
    )
    
    # test operations on a CUDA device if it's available
    # in the testing environment
    if torch.cuda.is_available():
        model = model.cuda()
        u_batch['input'] = u_batch['input'].cuda()
        d_batch['input'] = d_batch['input'].cuda()
        d_batch['output'] = d_batch['output'].cuda()
        
        # remove the old teacher because the CPU parameters
        # can't be updated with the CUDA parameters
        mixmatch.teacher = None
    
        loss_sup, loss_unsup, sup_outputs = mixmatch(
            model,
            unlabeled_sample=u_batch, 
            labeled_sample=d_batch,
        )
    
    return


def test_mixmatch_forward_with_confthresh():
    '''Perform a forward pass through the MixMatch objective
    using a dummy model and a pseudolabel confidence threshold.
    
    Notes
    -----
    Does not verify that computations are correct, merely that
    the forward pass completes without error.
    '''
    import scanpy as sc
    import scnym
    from scnym.dataprep import SingleCellDS
    torch.manual_seed(397)
    
    # load 10x human PBMC data as a sample
    adata = sc.datasets.pbmc3k()
    sc.pp.filter_cells(adata, min_counts=100)
    sc.pp.filter_genes(adata, min_cells=100)
    sc.pp.normalize_per_cell(adata, counts_per_cell_after=1e6)
    sc.pp.log1p(adata)
    
    # generate fake class labels
    adata.obs['class'] = np.random.randint(
        0,
        3,
        adata.shape[0],
    )
    
    # create dataloaders
    
    d_ds = SingleCellDS(
        X = adata.X.toarray(),
        y = np.array(adata.obs['class']),
    )
    u_ds = SingleCellDS(
        X = adata.X.toarray(),
        y = np.zeros(adata.shape[0]),
    )
    
    d_dl = torch.utils.data.DataLoader(d_ds, batch_size=16)
    u_dl = torch.utils.data.DataLoader(u_ds, batch_size=16)
    
    # draw a batch from each
    d_batch = next(iter(d_dl))
    u_batch = next(iter(u_dl))
    
    # initialize mixmatch using a dummy model
    unsup_criterion = torch.nn.MSELoss(reduction='none')
    sup_criterion = scnym.trainer.cross_entropy
    
    augment = scnym.dataprep.AUGMENTATION_SCHEMES['None']
    
    model = scnym.model.CellTypeCLF(
        n_layers=1,
        n_hidden=128,
        residual=False,
        init_dropout=0.0,
        n_genes=adata.shape[1],
        n_cell_types=3,
    )
    
    K = 1
    mixmatch = scnym.trainer.MixMatchLoss(
        unsup_criterion=unsup_criterion,
        sup_criterion=sup_criterion,
        augment=augment,
        alpha=0.3,
        n_augmentations=K,
        T=1.,
        pseudolabel_min_confidence=0.001,
    )

    # mixmatch forward pass
    loss_sup, loss_unsup, sup_outputs = mixmatch(
        model,
        unlabeled_sample=u_batch, 
        labeled_sample=d_batch,
    )
    ls_low_thresh = loss_sup.detach().item()
    lu_low_thresh = loss_unsup.detach().item()
    
    # double check that our running confidence scores
    # are working
    assert mixmatch.running_confidence_scores[0][0].size(0) == 16
    assert torch.sum(
        torch.cat(
            [x[0] for x in mixmatch.running_confidence_scores], 
            dim=0,
        ),
    ) > 0
    
    mixmatch = scnym.trainer.MixMatchLoss(
        unsup_criterion=unsup_criterion,
        sup_criterion=sup_criterion,
        augment=augment,
        alpha=0.3,
        n_augmentations=K,
        T=1.,
        pseudolabel_min_confidence=0.4,
    )

    # mixmatch forward pass
    loss_sup, loss_unsup, sup_outputs = mixmatch(
        model,
        unlabeled_sample=u_batch, 
        labeled_sample=d_batch,
    )
    lu_high_thresh = loss_unsup.detach().item()
    ls_high_thresh = loss_sup.detach().item()
    
    # assert that unsup losses decrease when we set higher thresholds
    print(f'Low  threshold unsup loss: {lu_low_thresh}')
    print(f'Low  threshold sup.  loss: {ls_low_thresh}')
    print(f'High threshold unsup loss: {lu_high_thresh}')
    print(f'High threshold sup.  loss: {ls_high_thresh}')    
    assert lu_high_thresh < lu_low_thresh
    # check that non-confident labels are being zero-ed out
    K = 1
    mixmatch = scnym.trainer.MixMatchLoss(
        unsup_criterion=unsup_criterion,
        sup_criterion=sup_criterion,
        augment=augment,
        alpha=0.3,
        n_augmentations=K,
        T=1.,
        pseudolabel_min_confidence=1.0,
    )

    # mixmatch forward pass
    loss_sup, loss_unsup, sup_outputs = mixmatch(
        model,
        unlabeled_sample=u_batch, 
        labeled_sample=d_batch,
    )
    msg = f'{loss_unsup.detach().item()} was not == 0.'
    assert loss_unsup.detach().item() == 0.
    print(f'Loss with confidence threshold == 1. was {loss_unsup.detach().item()}!')
            
    
    # test operations on a CUDA device if it's available
    # in the testing environment
    if torch.cuda.is_available():
        model = model.cuda()
        u_batch['input'] = u_batch['input'].cuda()
        d_batch['input'] = d_batch['input'].cuda()
        d_batch['output'] = d_batch['output'].cuda()
        
        # remove the old teacher because the CPU parameters
        # can't be updated with the CUDA parameters
        mixmatch.teacher = None        
    
        loss_sup, loss_unsup, sup_outputs = mixmatch(
            model,
            unlabeled_sample=u_batch, 
            labeled_sample=d_batch,
        )
    
    return


def test_mixmatch_forward_with_teacher_bn_runnning_stats():
    '''Perform a forward pass through the MixMatch objective
    using a dummy model and resampling augmentations
    
    Notes
    -----
    Does not verify that computations are correct, merely that
    the forward pass completes without error.
    '''
    
    import scanpy as sc
    import scnym
    from scnym.dataprep import SingleCellDS
    
    # load 10x human PBMC data as a sample
    adata = sc.datasets.pbmc3k()
    sc.pp.filter_cells(adata, min_counts=100)
    sc.pp.filter_genes(adata, min_cells=100)
    sc.pp.normalize_per_cell(adata, counts_per_cell_after=1e6)
    sc.pp.log1p(adata)
    
    # generate fake class labels
    adata.obs['class'] = np.random.randint(
        0,
        3,
        adata.shape[0],
    )
    
    # create dataloaders
    
    d_ds = SingleCellDS(
        X = adata.X.toarray(),
        y = np.array(adata.obs['class']),
    )
    u_ds = SingleCellDS(
        X = adata.X.toarray(),
        y = np.zeros(adata.shape[0]),
    )
    
    d_dl = torch.utils.data.DataLoader(d_ds, batch_size=16)
    u_dl = torch.utils.data.DataLoader(u_ds, batch_size=16)
    
    # draw a batch from each
    d_batch = next(iter(d_dl))
    u_batch = next(iter(u_dl))
    
    # initialize mixmatch using a dummy model
    unsup_criterion = torch.nn.MSELoss(reduction='none')
    sup_criterion = scnym.trainer.cross_entropy
    
    augment = scnym.dataprep.AUGMENTATION_SCHEMES['log1p_drop']
    
    model = scnym.model.CellTypeCLF(
        n_layers=1,
        n_hidden=128,
        residual=False,
        init_dropout=0.0,
        n_genes=adata.shape[1],
        n_cell_types=3,
    )
    
    K = 1
    mixmatch = scnym.trainer.MixMatchLoss(
        unsup_criterion=unsup_criterion,
        sup_criterion=sup_criterion,
        augment=augment,
        alpha=0.3,
        n_augmentations=K,
        teacher_bn_running_stats=False,
    )

    # mixmatch forward pass
    loss_sup, loss_unsup, sup_outputs = mixmatch(
        model,
        unlabeled_sample=u_batch, 
        labeled_sample=d_batch,
    )
    
    # test operations on a CUDA device if it's available
    # in the testing environment
    if torch.cuda.is_available():
        model = model.cuda()
        u_batch['input'] = u_batch['input'].cuda()
        d_batch['input'] = d_batch['input'].cuda()
        d_batch['output'] = d_batch['output'].cuda()
        
        # remove the old teacher because the CPU parameters
        # can't be updated with the CUDA parameters
        mixmatch.teacher = None
    
        loss_sup, loss_unsup, sup_outputs = mixmatch(
            model,
            unlabeled_sample=u_batch, 
            labeled_sample=d_batch,
        )
    
    return


def test_train_mixmatch():
    '''Train a few iterations of a mixmatch loss to ensure we're
    passing gradients properly'''
    import scanpy as sc
    import scnym
    from scnym.dataprep import SingleCellDS
    
    # load 10x human PBMC data as a sample
    adata = sc.datasets.pbmc3k()
    sc.pp.filter_cells(adata, min_counts=100)
    sc.pp.filter_genes(adata, min_cells=100)
    sc.pp.normalize_per_cell(adata, counts_per_cell_after=1e6)
    sc.pp.log1p(adata)
    
    # generate fake class labels
    adata.obs['class'] = np.random.randint(
        0,
        3,
        adata.shape[0],
    )
    
    # create dataloaders
    
    d_ds = SingleCellDS(
        X = adata.X.toarray(),
        y = np.array(adata.obs['class']),
    )
    u_ds = SingleCellDS(
        X = adata.X.toarray(),
        y = np.zeros(adata.shape[0]),
    )
    
    d_dl = torch.utils.data.DataLoader(d_ds, batch_size=16)
    u_dl = torch.utils.data.DataLoader(u_ds, batch_size=16)
    
    # draw a batch from each
    d_batch = next(iter(d_dl))
    u_batch = next(iter(u_dl))
    
    # initialize mixmatch using a dummy model
    unsup_criterion = torch.nn.MSELoss(reduction='none')
    sup_criterion = scnym.trainer.cross_entropy
    
    augment = scnym.dataprep.AUGMENTATION_SCHEMES['None']
    
    model = scnym.model.CellTypeCLF(
        n_layers=1,
        n_hidden=128,
        residual=False,
        init_dropout=0.0,
        n_genes=adata.shape[1],
        n_cell_types=3,
    )
    
    K = 1
    mixmatch = scnym.trainer.MixMatchLoss(
        unsup_criterion=unsup_criterion,
        sup_criterion=sup_criterion,
        augment=augment,
        alpha=0.3,
        n_augmentations=K,
        T=1.,
        pseudolabel_min_confidence=0.001,
    )

    # mixmatch forward pass
    loss_sup, loss_unsup, sup_outputs = mixmatch(
        model,
        unlabeled_sample=u_batch, 
        labeled_sample=d_batch,
    )
    init_sl = loss_sup.detach().item()
        
    # train some iterations to reduce sup loss
    optimizer = torch.optim.Adam(model.parameters())
    for i in range(50):
        optimizer.zero_grad()
        loss_sup, loss_unsup, sup_outputs = mixmatch(
            model,
            unlabeled_sample=u_batch, 
            labeled_sample=d_batch,
        )
        loss_sup.backward()
        
        # check that no gradients are passed to the teacher
        teacher_grad = [
            p.grad is None for p in mixmatch.teacher.parameters()
        ]
        assert sum(teacher_grad) == len(teacher_grad)
        
        optimizer.step()
    final_sl = loss_sup.detach().item()
    assert final_sl < init_sl
    
    # train some iterations to reduce unsup loss
    model = scnym.model.CellTypeCLF(
        n_layers=1,
        n_hidden=128,
        residual=False,
        init_dropout=0.0,
        n_genes=adata.shape[1],
        n_cell_types=3,
    )
    loss_sup, loss_unsup, sup_outputs = mixmatch(
        model,
        unlabeled_sample=u_batch, 
        labeled_sample=d_batch,
    )
    init_ul = loss_unsup.detach().item()    
    
    optimizer = torch.optim.Adam(model.parameters())
    for i in range(50):
        optimizer.zero_grad()
        loss_sup, loss_unsup, sup_outputs = mixmatch(
            model,
            unlabeled_sample=u_batch, 
            labeled_sample=d_batch,
        )
        loss_unsup.backward()
        optimizer.step()
    final_ul = loss_unsup.detach().item()
    assert final_ul < init_ul    
    
    return


def test_sharpen_labels():
    '''Test that label sharpening reduces entropy'''
    from scnym.trainer import sharpen_labels
    
    p = torch.FloatTensor([[0.1, 0.2, 0.7]])
    q = sharpen_labels(p, T=0.5)
    
    H = lambda p: -torch.sum(p * torch.log(p))
    
    assert H(p) > H(q)
    
    return


def test_sharpen_labels_t0():
    '''Test that label sharpening returns argmax one-hot
    labels when temperature is == 0.'''
    from scnym.trainer import sharpen_labels
    
    p = torch.FloatTensor([[0.1, 0.2, 0.7]])
    q = sharpen_labels(p, T=0.0)
    
    assert torch.all(
        q == torch.FloatTensor([0, 0, 1])
    )
    
    return


def test_weight_schedule():
    '''Test semi-supervised weight scheduling'''
    from scnym.trainer import ICLWeight
    
    # test ramping
    w = ICLWeight(
        ramp_epochs = 100,
        sigmoid = False,
        max_unsup_weight = 100.,
    )
    assert w(0) == 0.
    assert w(50) > 0.
    assert w(50) < 100.
    assert w(100) == 100.
    assert w(150) == 100.

    w = ICLWeight(
        ramp_epochs = 100,
        sigmoid = True,
        max_unsup_weight = 100.,
    )
    assert w(0) == np.exp(-5)*100.
    assert w(50) > 0.
    assert w(50) < 100.
    assert w(100) == 100.
    assert w(150) == 100.
    
    # test no ramp
    w = ICLWeight(
        ramp_epochs = 0,
        sigmoid = False,
        max_unsup_weight = 1.,
    )
    assert w(0) == 1.
    assert w(5) == 1.
    
    assert 0. < w(3)
    assert 1. >= w(3)
    
    w = ICLWeight(
        ramp_epochs = 0,
        sigmoid = True,
        max_unsup_weight = 1.,
    )
    assert w(0) == 1.
    assert w(5) == 1.    
    
    # test burn in
    w = ICLWeight(
        burn_in_epochs = 5,
        ramp_epochs = 5,
        sigmoid = False,
        max_unsup_weight = 1.,
    )
    assert w(0) == 0.
    assert w(4) == 0.
    assert w(10) == 1.
    
    # check that we only accept integer epochs
    with pytest.raises(TypeError):
        w('a')
    
    return


def main():
    test_samplemixup()
    test_mixmatch_forward_with_confthresh()
    return


if __name__ == '__main__':
    main()


import sys
if sys.version_info < (3,):
    sys.exit('scnym requires Python >= 3.6')
from pathlib import Path

from setuptools import setup, find_packages

try:
    from scnym import __author__, __email__
except ImportError:  # Deps not yet installed
    __author__ = __email__ = ''

setup(
    name='scnym',
    version='0.3.1',
    description="Semi supervised adversarial network networks for single cell classification",
    long_description="scNym uses the semi-supervised MixMatch framework and domain adversarial training to take advantage of information in both the labeled and unlabeled datasets.",
    url='http://github.com/calico/scnym',
    author=__author__,
    author_email=__email__,
    license='Apache',
    python_requires='>=3.6',
    install_requires=[
        l.strip() for l in
        Path('requirements.txt').read_text('utf-8').splitlines()
    ],
    packages=find_packages(),
    entry_points=dict(
        console_scripts=['scnym=scnym.main:main', 'scnym_ad=scnym.scnym_ad:main'],
    ),
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],
)



'''scNym model training from standard anndata objects'''
import anndata
import os
import os.path as osp
import uuid
import configargparse
import numpy as np
import pandas as pd

from .main import train_cv, train_all
from .model import CellTypeCLF
from .utils import build_classification_matrix
from sklearn.model_selection import StratifiedKFold


def make_parser():
    parser = configargparse.ArgParser(
        description='train an scNym cell type classification model.'
    )
    parser.add_argument(
        '--config',
        type=str,
        is_config_file=True,
        required=False,
        help='path to a configuration file.'
    )
    parser.add_argument(
        '--data',
        type=str,
        help='path to an h5ad [Cells, Features] object.'
    )
    parser.add_argument(
        '--groupby',
        type=str,
        help='categorical feature in `adata.obs` to use for classifier training.'
    )
    parser.add_argument(
        '--out_path',
        type=str,
        help='path for outputs.'
    )
    parser.add_argument(
        '--batch_size', 
        type=int, 
        default=256,
        help='batch size for training',
    )
    parser.add_argument(
        '--n_epochs', 
        type=int, 
        default=200,
        help='number of epochs for training',
    )
    parser.add_argument(
        '--init_dropout', 
        type=float, 
        default=0.0,
        help='initial dropout to perform on gene inputs',
    )
    parser.add_argument(
        '--n_hidden', 
        type=int, 
        default=128,                
        help='number of hidden units in the classifier',
    )
    parser.add_argument(
        '--n_layers', 
        type=int, 
        default=2,
        help='number of hidden layers in the model',
    )
    parser.add_argument(
        '--residual', 
        action='store_true',
        help='use residual layers in the model',
    )
    parser.add_argument(
        '--weight_decay', 
        type=float, 
        default=1e-5,
        help='weight decay applied by the optimizer',
    )
    parser.add_argument(
        '--weight_classes', 
        type=bool, 
        default=True,
        help='weight loss based on relative class abundance.',
    )
    parser.add_argument(
        '--mixup_alpha', 
        type=float, 
        default=None,
        help='alpha parameter for MixUp training. if set performs MixUp, otherwise does not.',
    )
    parser.add_argument(
        '--unlabeled_counts',
        type=str, 
        default=None,
        help='path to h5ad [Cells, Features] object of unlabeled data.',
    )
    parser.add_argument(
        '--unsup_max_weight', 
        type=float,
        default=2.,
        help='maximum weight for the unsupervised component of IC training.',
    )
    parser.add_argument(
        '--unsup_mean_teacher', 
        type=bool,
        default=True,
        help='use a mean teacher for IC training.',
    )
    parser.add_argument(
        '--cross_val_train', 
        action='store_true',
    )
    return parser


def main():
    parser = make_parser()
    args = parser.parse_args()
    
    adata = anndata.read_h5ad(args.data)
    print(f'{adata.shape[0]} cells, {adata.shape[1]} genes in the training data.')
    
    if args.groupby not in adata.obs:
        msg = f'{args.groupby} not in `adata.obs`'
        raise ValueError(msg)
        
    os.makedirs(args.out_path, exist_ok=True)
    
    
    if args.unlabeled_counts is None:
        unlabeled_counts = None
    else:
        # load unlabeled counts and build a matrix that follows
        # gene dimension ordering of the training data
        unlabeled_adata = anndata.read_h5ad(args.unlabeled_counts)
        unlabeled_counts = build_classification_matrix(
            X=unlabeled_adata.X if type(unlabeled_adata.X)==np.ndarray else unlabeled_adata.X.toarray(),
            model_genes=np.array(adata.var_names),
            sample_genes=np.array(unlabeled_adata.var_names),
        )
    
    X = adata.X if type(adata.X)==np.ndarray else adata.X.toarray()
    y = pd.Categorical(adata.obs[args.groupby]).codes
    
    model_params = {
        'n_hidden' : args.n_hidden,
        'residual' : args.residual,
        'n_layers' : args.n_layers,
        'init_dropout' : args.init_dropout,        
    }
    
    if args.cross_val_train:
        kf = StratifiedKFold(n_splits=5, shuffle=True)
        fold_indices = list(kf.split(X, y))

        fold_eval_acc, fold_eval_losses = train_cv(
            X=X,
            y=y,
            batch_size=args.batch_size,
            n_epochs=args.n_epochs,
            weight_decay=args.weight_decay,
            ModelClass=CellTypeCLF,
            fold_indices=fold_indices,
            out_path=args.out_path,
            n_genes=adata.shape[1],
            mixup_alpha=args.mixup_alpha,
            unlabeled_counts=unlabeled_counts,
            unsup_max_weight=args.unsup_max_weight,
            unsup_mean_teacher=args.unsup_mean_teacher,
            weighted_classes=args.weight_classes,
            **model_params,
        )
        np.savetxt(
            osp.join(args.out_path, 'fold_eval_losses.csv',),
            fold_eval_losses,
        )
        np.savetxt(
            osp.join(args.out_path, 'fold_eval_acc.csv',),
            fold_eval_acc,
        )
    
    val_loss, val_acc = train_all(
        X=X,
        y=y,
        batch_size=args.batch_size,
        n_epochs=args.n_epochs,
        weight_decay=args.weight_decay,
        ModelClass=CellTypeCLF,
        out_path=args.out_path,
        n_genes=adata.shape[1],
        mixup_alpha=args.mixup_alpha,
        unlabeled_counts=unlabeled_counts,
        unsup_max_weight=args.unsup_max_weight,
        unsup_mean_teacher=args.unsup_mean_teacher,
        weighted_classes=args.weight_classes,
        **model_params,
    )
    print(f'Final validation loss: {val_loss:08}')
    print(f'Final validation acc : {val_acc:08}')
    
    # get exp id
    exp_id = uuid.uuid4()
    res = pd.DataFrame(
        {'val_acc': val_acc, 'val_loss': val_loss},
        index=[exp_id],
    ).to_csv(
        osp.join(
            args.out_path,
            'all_data_val_results.csv',
        )
    )
    return
